package Features;

use Document;

# Per les dates
use Date::Manip;

use strict;

# Optimitzacio
use Algorithm::Munkres;

##############################################################################
# globals

use Globals qw($PATH_RESOURCES $WNLIB $LANG);
use lib $WNLIB;
use WordNet::QueryData;

our %LIST_COUNTRIES;
our %LIST_CAPITALS;
our %LIST_GENTILIC;
our %LIST_NICKNAMES;
our %LIST_DS;

my $wn = undef;

#
# Feature Functions for mentions
#

# Carrega Wordnet
sub InitWN
{
	if (!defined($wn)) {
		$wn = WordNet::QueryData->new;
	}
}

# Retorna el text d'un markable. El text son el contingut de cadascun dels
# tokens. Es separen per espais o sigui que la puntuacio pot no quedar gaire be
sub GetString
{
	my $markable = shift;
	return $markable->string;
}

sub MentionType
{
	my $mencio = shift;
	return $mencio->type;
}

sub Is3Person
{
	my $mencio = shift;
	
	return 1 if ($mencio->person == 3);
	return 0;
}

sub MorPerson
{
	my $mencio = shift;
	return $mencio->person;
}

sub IsPronoun
{
	my $mencio = shift;
	
	return $mencio->isPronoun;
}

sub IsPossessive
{
	my $mencio = shift;
	
	return $mencio->isPossessive;
}

sub InQuotes
{
	my $mencio = shift;
	return $mencio->inQuotes;
}

sub FirstInSentence
{
	my $mencio = shift;
	return $mencio->firstInSentence;
}

sub StringMatch
{
	my ($mencioA, $mencioB) = @_;
	my $doc = $mencioA->document;
	
	my $txtA = $mencioA->string;
	my $txtB = $mencioB->string;

	# eliminates 'the'
	foreach my $txt ($txtA, $txtB) {
		$txt = TreuDTs($txt);
	}

	# No case sensitive in first character
	$txtA = lc(substr($txtA, 0, 1)) . substr($txtA, 1);
	$txtB = lc(substr($txtB, 0, 1)) . substr($txtB, 1);
	
	my $r = ($txtA eq $txtB) ? 1 : 0;
	
	return $r;
}

sub HeadMatch
{
	my ($mencioA, $mencioB) = @_;
	my $doc = $mencioA->document;
	
	my $txtA = $mencioA->head_str;
	my $txtB = $mencioB->head_str;
		
	my $r = (lc($txtA) eq lc($txtB)) ? 1 : 0;
	
	return $r;
}

sub TermMatch
{
	my ($mencioA, $mencioB) = @_;
	my $doc = $mencioA->document;
	
	my $txtA = $mencioA->term;
	my $txtB = $mencioB->term;
	
	my $r = (lc($txtA) eq lc($txtB)) ? 1 : 0;
	
	return $r;
}

# Mirem que no hi hagi cap mencio anidada per sobre
sub IsMaximalNP
{
	my $mencio = shift;
	return $mencio->isMaximalNP;
}

sub GetMaximalNP
{
	my $mencio = shift;
	return $mencio->GetMaximalNP;
}

sub MaximalNP
{
	my ($mencioA, $mencioB) = @_;
	my $doc = $mencioA->document;
	
	my $maxA = $mencioA->GetMaximalNP;
	my $maxB = $mencioB->GetMaximalNP;
	my $r = ($maxA->{id} eq $maxB->{id}) ? 1 : 0;
	
	return $r;
}

sub IsEmbeddedNoun
{
	my $mencio = shift;
	
	if ($mencio->isNoun && !$mencio->isMaximalNP) {
		return 1;
	}
	return 0;
}


# SIMPLE: Dues mencions son "appositive" quan estan separades per una coma
sub Appositive
{
	my ($mencioA, $mencioB) = @_;
	my $doc = $mencioA->document;
	
	my $r = 0;
	if (DistSentences($mencioA, $mencioB) == 0) {
	
		# We assure that mencioA goes before mencioB in the document
		if ($mencioA->{span_ini} > $mencioB->{span_ini}) {
			($mencioA, $mencioB) = ($mencioB, $mencioA);
		}
		
		if ($mencioA->{span_fi}+2 >= $mencioB->{span_ini}) {
			my $tokens = $doc->tokens;
			if ($tokens->[$mencioA->{span_fi}+1]->{content} eq ',') {
				$r = 1;
			}
		}
	}
	
	return $r;
}

sub IsComposite
{
	my $mencio = shift;
	return $mencio->isComposite;
}

sub DistSentences
{
	my ($mencioA, $mencioB, $limit) = @_;
	my $senA = $mencioA->sentence;
	my $senB = $mencioB->sentence;
	my $dist = abs($senA->{index} - $senB->{index});
	
	# by default, the limit distance is 3
	$limit = 3 if (!defined($limit));
	
	# limit = 0 means no limit
	if ($limit > 0) {
		if ($dist > $limit) {
			$dist = $limit;
		}
	}
	
	return $dist;
}

sub DistMentions
{
	my ($mencioA, $mencioB, $limit) = @_;
	
	# Confiem en que les IDs estan per ordre d'aparició
	my $dist = abs($mencioA->index - $mencioB->index);
	
	# by default, the limit distance is 10
	$limit = 10 if (!defined($limit));
	
	# limit = 0 means no limit
	if ($limit > 0) {
		if ($dist > $limit) {
			$dist = $limit;
		}
	}
	
	return $dist;
}


sub DefNP
{
	my $mencio = shift;
	return $mencio->isDefNP;
}

sub IndefNP
{
	my $mencio = shift;
	return $mencio->isIndefNP;
}


sub DemNP
{
	my $mencio = shift;
	return $mencio->isDemNP;
}


# Si la mencio es plural retorna 1
# si es singular o no pot confirmar-ho, retorna 0
sub IsPlural
{
	my $mencio = shift;
	return $mencio->isPlural;
}


sub NumberMatch
{
	my ($mencioA, $mencioB) = @_;
	my $doc = $mencioA->document;
	
	my $r = 2; # unknown by default
	if ($mencioA->isPlural != 2 && $mencioA->isPlural != 2) {
		$r = ($mencioA->isPlural == $mencioB->isPlural) ? 1 : 0;
	}
	
	return $r;
}

sub SemanticClass
{
	my $mencio = shift;
	return $mencio->semanticClass;
}

sub EnglobaClass
{
	my ($synsetA, $synsetB) = @_;
	
	return 1 if ($synsetA ne 'unknown' && $synsetA eq $synsetB);
	return 1 if ($synsetA =~ /male/ && ($synsetB eq 'person#n#1'));
	return 1 if ($synsetB =~ /male/ && ($synsetA eq 'person#n#1'));
	return 1 if ($synsetA eq 'gpe' && ($synsetB =~ /^(location|organization)\#/));
	return 1 if ($synsetB eq 'gpe' && ($synsetA =~ /^(location|organization)\#/));
	return 0;
}

sub SemanticClassMatch
{
	my ($mencioA, $mencioB) = @_;
	my $doc = $mencioA->document;
	
	my $semA = $mencioA->semanticClass;
	my $semB = $mencioB->semanticClass;
	my $r = 2; #unknown by default
	if ($semA ne 'unknown' && ($semB ne 'unknown')) {
		if ($semA eq $semB) {
			$r = 1;
		}
		elsif (EnglobaClass($semA, $semB)) {
			$r = 1;
		}
		else {
			$r = 0;
		}
	}
	elsif (lc($mencioA->head_str) eq lc($mencioB->head_str)) {
		$r = 1;
	}

	return $r;
}


sub IsPerson
{
	my $mencio = shift;
	return $mencio->isPerson;
}

sub IsOrganization
{
	my $mencio = shift;
	return ($mencio->semanticClass =~ /^organization/) ? 1 : 0;
}

sub IsLocation
{
	my $mencio = shift;
	return ($mencio->semanticClass =~ /^location/) ? 1 : 0;
}

sub Animacy
{
	my ($mencioA, $mencioB) = @_;
	my $doc = $mencioA->document;
	my $r = 0; #default
	if ($mencioA->isPerson && $mencioB->isPerson) {
		$r = 1;
	}
	elsif (!$mencioA->isPerson && !$mencioB->isPerson) {
		$r = 1;
	}
	
	return $r;
}

sub IsReflexive
{
	my $mencio = shift;
	return $mencio->isReflexive;
}

sub LoadNicknames
{
	if (keys(%LIST_NICKNAMES)) {
		return \%LIST_NICKNAMES;
	}
	
	open (L, "$PATH_RESOURCES/llista-nicknames.txt") || die "Can not open $PATH_RESOURCES/llista-nicknames.txt";
	my $count = 0;
	while (my $l = <L>) {
		chomp($l);
		my @names = split(/,/, $l);
		foreach my $name (@names) {
			push(@{$LIST_NICKNAMES{$name}}, $count);
		}
		$count++;
	}
	close (L);
	
	return \%LIST_NICKNAMES;
}

sub NickNameMatch
{
	my ($str1, $str2) = @_;
	
	my $list = LoadNicknames;
	if (defined($list->{$str1}) && defined($list->{$str2})) {
		foreach my $a (@{$list->{$str1}}) {
			foreach my $b (@{$list->{$str2}}) {
				return 1 if ($a == $b);
			}
		}
		return 0;
	}
	
	return 2;
}

sub LoadCountries
{
	if (keys(%LIST_COUNTRIES)) {
		return (\%LIST_COUNTRIES, \%LIST_CAPITALS, \%LIST_GENTILIC);
	}
	
	open (CAP, "$PATH_RESOURCES/capitals.txt") || die "Can not open capitals.txt";
	while (my $l = <CAP>) {
		chomp($l);
		my ($pais, $capitals) = split(/\s*\-\s*/, $l);
		my @ciutats = split(/\s*\;\s*/, $l);
		foreach my $p ($pais, @ciutats) {
			$p =~ s/^\s+//;
			$p =~ s/\s+$//;
			$p =~ s/\s+/ /;
		}
		$LIST_COUNTRIES{$pais} = \@ciutats;
		foreach my $c (@ciutats) {
			$LIST_CAPITALS{$c} = $pais;
		}
	}
	close (CAP);
	
	open (GEN, "$PATH_RESOURCES/gentilicis.txt") || die "Can not open gentilicis.txt";
	while (my $l = <GEN>) {
		chomp($l);
		my ($gentilici, $pais) = split(/\s+/, $l);
		$LIST_GENTILIC{$gentilici} = $pais;
	}
	close (GEN);
	
	return (\%LIST_COUNTRIES, \%LIST_CAPITALS, \%LIST_GENTILIC);
}

sub GeoMatch
{
	my ($mencioA, $mencioB) = @_;
	
	return 2 if (!$mencioA->isEntity || !$mencioB->isEntity);
	if ($mencioA->semanticClass !~ /location/ || $mencioB->semanticClass !~ /location/) {
		my ($countries, $capitals, $gentilicis) = LoadCountries;
		foreach my $i (0,1) {
			my ($a, $b) = $i ? ($mencioA, $mencioB) : ($mencioB, $mencioA);
			if ($a->semanticClass =~ /location/) {
				my $entA = $a->GetEntity;
				my $strA = $entA->string;
				my $strB = $b->head_str;
				if (defined($countries->{$strA})) {
					if (defined($gentilicis->{lc($strB)})) {
						if ($gentilicis->{lc($strB)} eq $strA) {
							return 1;
						}
						else {
							return 0;
						}
					}
				}
			}
		}
		return 2;
	}
	my $entA = $mencioA->GetEntity;
	my $entB = $mencioB->GetEntity;
	my $strA = $entA->string;
	my $strB = $entB->string;
	
	return 1 if ($strA eq $strB);
	
	my ($countries, $capitals, $gentilicis) = LoadCountries;
	
# 	return 1 if (defined($countries{$strA}) && $countries{$strA} eq $strB);
# 	return 1 if (defined($countries{$strB}) && $countries{$strB} eq $strA);
	return 1 if (defined($capitals->{$strA}) && $capitals->{$strA} eq $strB);
	return 1 if (defined($capitals->{$strB}) && $capitals->{$strB} eq $strA);
	
	
	return 0 if ((defined($countries->{$strA}) || defined($capitals->{$strA})) &&
	(defined($countries->{$strB}) || defined($capitals->{$strB})));
	
	return 2;
}

sub GenderMatch
{
	my ($mencioA, $mencioB) = @_;
	my $doc = $mencioA->document;
	my $gA = $mencioA->gender;
	my $gB = $mencioB->gender;
	my $r;
	if ($gA eq 'u' || $gB eq 'u') {
		$r = 2;
	}
	else {
		$r = ($gA eq $gB) ? 1 : 0;
	}

	return $r;
}

sub Nested
{
	my ($mencioA, $mencioB) = @_;
	
	if (($mencioA->{span_ini} <= $mencioB->{span_ini} &&
		$mencioA->{span_fi} >= $mencioB->{span_fi}) ||
		($mencioA->{span_ini} >= $mencioB->{span_ini} &&
		$mencioA->{span_fi} <= $mencioB->{span_fi})) {
		return 1;
	}
	
	return 0;
}


sub ParentNP
{
	my $mencio = shift;
	return $mencio->parentNP;
}

# Retorna 1 si A C-Command B (Binding Theory)
# 1. A no domina B
# 2. B no domina A
# 3. La primera branca que domina A tambe domina B
sub C_Commands
{
	my ($mencioA, $mencioB) = @_;
	my $doc = $mencioA->document;
	my $r;
	if (DistSentences($mencioA, $mencioB) > 0) {
		$r = 0;
	}
	elsif (Nested($mencioA, $mencioB)) {
		$r = 0;
	}
	else {
		# S'ha de trobar el NP inmediatament per sobre de A
		my $pare = ParentNP($mencioA);
		$r = (Nested($pare, $mencioB)) ? 1 : 0;
	}
	
	return $r;
}

# Retorna 1 si les dues mencions compleixen les condicions positives de la
# Binding Theory
sub BindingPos
{
	my ($mencioA, $mencioB) = @_;
	my $doc = $mencioA->document;

	my $r = 0; #default
	if ($mencioB->isReflexive && C_Commands($mencioA, $mencioB) &&
		AGREEMENT($mencioA, $mencioB) == 1 && Animacy($mencioA, $mencioB)) {
		my $ta = $mencioA->string;
		my $tb = $mencioB->string;
		$r = 1;
	}

	return $r;
}

# Retorna 1 si les dues mencions compleixen les condicions negatives de la
# Binding Theory
sub BindingNeg
{
	my ($mencioA, $mencioB) = @_;
	my $doc = $mencioA->document;

	my $r = 0;
	if ($mencioB->isPronoun && !$mencioB->isReflexive &&
	C_Commands($mencioA, $mencioB)) {
		$r = 1;
	}
	elsif (!$mencioB->isPronoun && C_Commands($mencioA, $mencioB)) {
		$r = 1;
	}
	
	return $r;
}

# Torna 1 si es un nom propi, sino 0
sub ProperName
{
	my $mencio = shift;
	return $mencio->isProperName;
}

sub IsNoun
{
	my $mencio = shift;
	return $mencio->isNoun;
}


sub IsEntity
{
	my $mencio = shift;
	return $mencio->isEntity;
}


sub GetEntity
{
	my $mencio = shift;
	return $mencio->GetEntity;
}

sub SyntaxFunction
{
	my $mencio = shift;
	return $mencio->syntaxFunction;
}

sub ProperNameMatch
{
	my ($mencioA, $mencioB) = @_;
	my $doc = $mencioA->document;

	my $r = 2; # default
	if ($mencioA->isProperName && $mencioB->isProperName) {
		$r = 1;
	}
	elsif (!$mencioA->isProperName && !$mencioB->isProperName) {
		$r = 0;
	}

	return $r;
}

# minimal element of a list
#
sub min
{
	my @list = @{$_[0]};
	my $min = $list[0];
	
	foreach my $i (@list)
	{
		$min = $i if ($i < $min);
	}
	
	return $min;
}

sub levenshtein
{
	# $s1 and $s2 are the two strings
	# $len1 and $len2 are their respective lengths
	#
	my ($s1, $s2) = @_;
	my ($len1, $len2) = (length $s1, length $s2);
	
	# If one of the strings is empty, the distance is the length
	# of the other string
	#
	return $len2 if ($len1 == 0);
	return $len1 if ($len2 == 0);
	
	my %mat;
	
	# Init the distance matrix
	#
	# The first row to 0..$len1
	# The first column to 0..$len2
	# The rest to 0
	#
	# The first row and column are initialized so to denote distance
	# from the empty string
	#
	for (my $i = 0; $i <= $len1; ++$i)
	{
		for (my $j = 0; $j <= $len2; ++$j)
		{
			$mat{$i}{$j} = 0;
			$mat{0}{$j} = $j;
		}
		
		$mat{$i}{0} = $i;
	}
	
	# Some char-by-char processing is ahead, so prepare
	# array of chars from the strings
	#
	my @ar1 = split(//, $s1);
	my @ar2 = split(//, $s2);
	
	for (my $i = 1; $i <= $len1; ++$i)
	{
		for (my $j = 1; $j <= $len2; ++$j)
		{
			# Set the cost to 1 iff the ith char of $s1
			# equals the jth of $s2
			# 
			# Denotes a substitution cost. When the char are equal
			# there is no need to substitute, so the cost is 0
			#
			my $cost = ($ar1[$i-1] eq $ar2[$j-1]) ? 0 : 1;
			
			# Cell $mat{$i}{$j} equals the minimum of:
			#
			# - The cell immediately above plus 1
			# - The cell immediately to the left plus 1
			# - The cell diagonally above and to the left plus the cost
			#
			# We can either insert a new char, delete a char or
			# substitute an existing char (with an associated cost)
			#
			$mat{$i}{$j} = min([$mat{$i-1}{$j} + 1,
									  $mat{$i}{$j-1} + 1,
									  $mat{$i-1}{$j-1} + $cost]);
		}
	}
	
	# Finally, the Levenshtein distance equals the rightmost bottom cell
	# of the matrix
	#
	# Note that $mat{$x}{$y} denotes the distance between the substrings
	# 1..$x and 1..$y
	#
	return $mat{$len1}{$len2};
}


# TODO: treure angles hardcoded!!
sub TreuDTs
{
	my $txt = shift;
	$txt =~ s/^\b(a|an|the|this|these|that|those)\b//i;
	# Catala
# 	$txt =~ s/^\b(un|una|uns|unes|el|la|l\'?|els|les|aquest|aquesta|aquests|aquestes|aquell|aquella|aquells|aquelles)\b//i;
	# Castella
# 	$txt =~ s/^\b(un|una|unos|unas|el|la|los|las|este|esta|estos|estas|eselesa|esos|esas|aquel|aquella|aquellos|aquellas)\b//i;
	
	$txt =~ s/\W//g;
	$txt =~ s/\s+/ /g;
	$txt =~ s/^\s+//g;
	$txt =~ s/\s+$//g;
	return $txt;
}


sub Words
{
	my $str = shift;
	my @paraules;
	# Acronim
	if (($str =~ /^[A-Z]{2,}$/) || ($str =~ /^([A-Z]\.)+$/)) {
		$str =~ s/\.//g;
		@paraules = split(//, $str);
	}
	else {
		@paraules = split(/\s+/, $str);
	}
	return @paraules;
}

sub LexDist
{
	my ($sa, $sb) = @_;
	
	return 0 if ($sa eq $sb);
	
	my $la = length($sa);
	my $lb = length($sb);
	
	# Acronim?
	if ($la == 1) {
		return 0.1 if ($sb =~ /^\Q$sa/);
		return 0.2 if ($sb =~ /^\Q$sa/i);
	}
	elsif ($lb == 1) {
		return 0.1 if ($sa =~ /^\Q$sb/);
		return 0.2 if ($sa =~ /^\Q$sb/i);
	}
	
	if (NickNameMatch($sa, $sb) == 1) {
		return 0.1;
	}
	elsif (NickNameMatch($sa, $sb) == 0) {
		return 1;
	}
	
	my $ml = ($la > $lb) ? $la : $lb;
	return levenshtein($sa, $sb) / $ml;
}

sub IsAcronim
{
	my $str = shift;
	if (($str =~ /^[A-Z]{2,}$/) || ($str =~ /^([A-Z]\.)+$/)) {
		return 1;
	}
	return 0;
}

# Funcio que mira si dos strings podrien ser alias. Va be per cridar-la desde fora
sub AliasLexic
{
	my ($stra, $strb) = @_;
	
	return 1 if (lc($stra) eq lc($strb));
	
	# Separem en paraules
	my @wordsa = Words($stra);
	my @wordsb = Words($strb);
	
	my @curt;
	my @llarg;
	if (@wordsa < @wordsb) {
		@curt = @wordsa;
		@llarg = @wordsb;
	}
	else {
		@curt = @wordsb;
		@llarg = @wordsa;
	}
	
	my @dist;
	for (my $i = 0; $i < @curt; $i++) {
		for (my $j = 0; $j < @llarg; $j++) {
			$dist[$i][$j] = LexDist($curt[$i], $llarg[$j]);
		}
	}
	
	my @out;
	# Munkres algorithm
	assign(\@dist, \@out);
	
	for (my $i = 0; $i < @curt; $i++) {
		return 0 if ($dist[$i][$out[$i]] > 0.2); # <== 0.3 pot ser massa generos, abans 0.2
	}
	
	return 1;
}

sub Alias
{
	my ($ma, $mb) = @_;
	
	if ($ma->isPronoun || $mb->isPronoun) {
		return 2;
	}
	
	# Un dels 2 es named entity o nom propi
	if ($ma->isEntity || $mb->isEntity || $ma->isProperName || $mb->isProperName) {
	
		my $stra = ($ma->isEntity) ? $ma->head_str : $ma->stringClean;
		my $strb = ($mb->isEntity) ? $mb->head_str : $mb->stringClean;
		
		return 1 if (lc($stra) eq lc($strb));
		
		my $geo = GeoMatch($ma, $mb);
		return $geo if ($geo != 2);
		
		# Separem en paraules
		my @wordsa = Words($stra);
		my @wordsb = Words($strb);
		
		# Els acronims han de coincidir: Nombre de lletres de l'acronim amb nombre de
		# paraules que comencen amb majuscula
		if (IsAcronim($stra)) {
			return 0 if (@wordsa > @wordsb);
		}
		if (IsAcronim($strb)) {
			return 0 if (@wordsa < @wordsb);
		}
		
		#alias lexic
		my @curt;
		my @llarg;
		if (@wordsa < @wordsb) {
			@curt = @wordsa;
			@llarg = @wordsb;
		}
		else {
			@curt = @wordsb;
			@llarg = @wordsa;
		}
		
		my @dist;
		for (my $i = 0; $i < @curt; $i++) {
			for (my $j = 0; $j < @llarg; $j++) {
				$dist[$i][$j] = LexDist($curt[$i], $llarg[$j]);
			}
		}
		
		my @out;
		# Munkres algorithm
		assign(\@dist, \@out);
		
		for (my $i = 0; $i < @curt; $i++) {
			return 0 if ($dist[$i][$out[$i]] > 0.2);
		}
		
		return 1;
	}
	
	return 2;
}


sub PRO_STR
{
	my ($mencioA, $mencioB) = @_;
	my $doc = $mencioA->document;
	my $r = 0;
	if ($mencioA->isPronoun && $mencioB->isPronoun &&
	StringMatch($mencioA, $mencioB)) {
		$r = 1;
	}
	
	return $r;
}

sub PN_STR
{
	my ($mencioA, $mencioB) = @_;
	my $doc = $mencioA->document;
	my $r = 0;
	if ($mencioA->isProperName && $mencioB->isProperName &&
		StringMatch($mencioA, $mencioB)) {
		$r = 1;
	}
	
	return $r;
}

sub SOON_STR_NONPRO
{
	my ($mencioA, $mencioB) = @_;
	my $doc = $mencioA->document;
	my $r = 0;
	if (!$mencioA->isPronoun && !$mencioB->isPronoun &&
	StringMatch($mencioA, $mencioB)) {
		$r = 1;
	}

	return $r;
}

sub BOTH_PRONOUNS
{
	my ($mencioA, $mencioB) = @_;
	my $doc = $mencioA->document;
	my $r = 2; #default unknown
	if ($mencioA->isPronoun && $mencioB->isPronoun) {
		$r = 1;
	}
	elsif (!$mencioA->isPronoun && !$mencioB->isPronoun) {
		$r = 0;
	}
	
	return $r;
}

sub AGREEMENT
{
	my ($mencioA, $mencioB) = @_;
	my $doc = $mencioA->document;
	my $r = 2;
	if (GenderMatch($mencioA, $mencioB) == 1 && 
	NumberMatch($mencioA, $mencioB) == 1) {
		$r = 1;
	}
	elsif (GenderMatch($mencioA, $mencioB) == 0 && 
	NumberMatch($mencioA, $mencioB) == 0) {
		$r = 0;
	}
	
	return $r;
}

sub ClosestAgreement
{
	my ($mencioA, $mencioB) = @_;
	my $doc = $mencioA->document;
	if (AGREEMENT($mencioA, $mencioB) == 1) {
		my $mentions = $doc->GetMentions($mencioA->trueMention);
		for (my $i = $mencioA->index + 1; $i < $mencioB->index; $i++) {
			return 0 if (AGREEMENT($mentions->[$i], $mencioB) == 1);
		}
		return 1;
	}
	
	return 2;
}

sub DistPhrases
{
	my ($mencioA, $mencioB) = @_;
	my $doc = $mencioA->document;
	my $r = -1;
	if (DistSentences($mencioA, $mencioB) == 0) {
		my $predsA = $mencioA->srl_apreds;
		my $predsB = $mencioB->srl_apreds;
		my $fA = 0;
		my $fB = 0;
		for (my $i=0; $i < @$predsA; $i++) {
			if (defined($predsA->[$i])) {
				$fA = $i;
			}
			if (defined($predsB->[$i])) {
				$fB = $i;
			}
		}
		$r = abs($fB - $fA);
	}
	
	return $r;
}

sub FindMarkers
{
	my $context = shift;
	my $r = 0;
	
	if (keys(%LIST_DS) == 0) {
		open (F, "$PATH_RESOURCES/discourse-markers.txt") || die "Error opening $PATH_RESOURCES/discourse-markers.txt: $!";
		while (my $l = <F>) {
			chomp($l);
			my ($type, $language, $val) = split(/:/, $l);
			if ($language eq $LANG) {
				$LIST_DS{$type} = $val;
			}
		}
		close (F);
	}
	
	if ($context =~ / ($LIST_DS{revision}) /) { $r = 1; }
	elsif ($context =~ / ($LIST_DS{cause}) /) { $r = 2; }
	elsif ($context =~ / ($LIST_DS{equality}) /) { $r = 3; }
	elsif ($context =~ / ($LIST_DS{context}) /) { $r = 4; }
	
	return $r;
}

# condicions: 
# Interna: les mencions estan a dues "phrases" seguides, i entre mig hi
#          ha un marcador de discurs
# Externa: TODO caldria mirar els marcadors de discurs a l'inici de
#          sentencia o paragraf
# Valors: 0 = res
#         1 = revision
#         2 = cause
#         3 = equality
#         4 = context
sub DiscourseLinked
{
	my ($mencioA, $mencioB) = @_;
	my $doc = $mencioA->document;

	my $r = 0;
	if (DistPhrases($mencioA, $mencioB) == 0 || DistPhrases($mencioA, $mencioB) == 1) {
		my $tokens = $doc->tokens;
		
		my $context = ' ';
		for (my $i = $mencioA->{span_fi} + 1; $i < $mencioB->{span_ini}; $i++) {
			$context .= lc($tokens->[$i]->content) . ' ';
		}
		
		$r = FindMarkers($context);
		
	}
	
	return $r;
}

# Mira si una mencio te un marcador de discurs en el seu context esquerra
sub HasDiscourseMarker
{
	my ($mencioA) = @_;
	if (!defined($mencioA->{_hasDiscourseMarker})) {
		my $doc = $mencioA->document;
		my $tokens = $doc->tokens;
		my $sentence = $mencioA->sentence;
		my $context = ' ';
		for (my $i = $sentence->{span_ini}; $i < $mencioA->{span_ini}; $i++) {
			$context .= lc($tokens->[$i]->content) . ' ';
		}
		$mencioA->{_hasDiscourseMarker} = FindMarkers($context);
	}
	return $mencioA->{_hasDiscourseMarker};
}


# Equivalent a DIST_PHR = 0
sub SRL_SameVerb
{
	my ($mencioA, $mencioB) = @_;
	my $doc = $mencioA->document;
	my $r = 0;
	if (DistSentences($mencioA, $mencioB) == 0) {
		my $predsA = $mencioA->srl_apreds;
		my $predsB = $mencioB->srl_apreds;
		for (my $i=0; $i < @$predsA; $i++) {
			if (defined($predsA->[$i]) && defined($predsB->[$i])) {
				$r = 1;
				last;
			}
		}
	}
	
	return $r;
}

sub SRL_Arg
{
	my $mencio = shift;
	return $mencio->srl_arg;
}

sub SRL_Role
{
	my $mencio = shift;
	return $mencio->srl_role;
}

sub SRL_SameArg
{
	my ($mencioA, $mencioB) = @_;
	my $doc = $mencioA->document;
	my $r = 0;
	if ($mencioA->srl_arg eq $mencioB->srl_arg) {
		$r = 1;
	}
	
	return $r;
}

sub SRL_SameRole
{
	my ($mencioA, $mencioB) = @_;
	my $doc = $mencioA->document;
	my $r = 0;
	if ($mencioA->srl_role eq $mencioB->srl_role) {
		$r = 1;
	}
	
	return $r;
}

sub HeadSynonym
{
	my ($mencioA, $mencioB) = @_;
	my $doc = $mencioA->document;
	
	return 2 if ($mencioA->type ne 's' || $mencioB->type ne 's');
	
	return 1 if (HeadMatch($mencioA, $mencioB) == 1);
	
	my $r = 0;
	
	my $txtA = $mencioA->head_str;
	my $txtB = $mencioB->head_str;
	
	$txtA =~ s/\#//og;
	$txtB =~ s/\#//og;
	
	InitWN;
	my @a = $wn->querySense("$txtA#n#1", 'syns');
	
	foreach my $sino (@a) {
		if ($sino =~ /^\Q$txtB\E\#/) {
			$r = 1;
			last;
		}
	}
	
	
	return $r;
}

# Retorna 1 si mA i mB estan separades pel verb "to be"
# HArdcoded: English
sub VerbIs
{
	my ($mencioA, $mencioB) = @_;
	my $doc = $mencioA->document;

	my $r = 0; #default
	my $doc = $mencioA->document;
	my $tokens = $doc->tokens;
	my $pos = $doc->GetLevel('pos');
	if (DistSentences($mencioA, $mencioB) == 0
		&& $mencioA->isMaximalNP && $mencioB->isMaximalNP) {
		
		# Ens assegurem de que A va abans de B
		if ($mencioA->{span_ini} > $mencioB->{span_ini}) {
			($mencioA, $mencioB) = ($mencioB, $mencioA);
		}
		
		if ($mencioB->{span_ini} - $mencioA->{span_fi} < 5) {
			my $ok = 0;
			for (my $i = $mencioA->{span_fi} + 1; $i < $mencioB->{span_ini}; $i++) {
				my $t = $tokens->[$i];
				if ($LANG eq 'en') {
					if ($pos->[$t->{pos_level}]->{tag} =~ /^VB/ &&
						$t->content =~ /^(is|was|were|a?m|are)$/) {
						$ok = 1;
					}
					elsif ($ok && $pos->[$t->{pos_level}]->{tag} =~ /^VB/) {
						$ok = 0;
					}
				}
			}
			
			$r = 1 if ($ok);
		}
	}
	
	return $r;
}

# Incompatibles per significat: [[his] father]
sub Incompatibles
{
	my ($mencioA, $mencioB) = @_;
	my $doc = $mencioA->document;
	my $r = 0; #default
	if (&Features::Nested($mencioA, $mencioB) == 1) {
		if (($mencioA->isPossessive || $mencioB->isPossessive)
		&& ($mencioA->isPronoun == 0 || $mencioB->isPronoun == 0)) {
			$r = 1;
		}
	}
	
	return 0;
}

sub SameHeadTerm
{
	my ($mencioA) = @_;
	
	return ($mencioA->head_str eq $mencioA->term) ? 1 : 0;
}

sub SameSpeaker
{
	my ($mencioA, $mencioB) = @_;
	my $spA = $mencioA->speaker;
	my $spB = $mencioB->speaker;
	if ($spA eq 'uknown' || $spB eq 'uknown') {
		return 2;
	}
	return (lc($spA) eq lc($spB)) ? 1 : 0;
}

# El nom 
sub SpeakerMatch
{
	my ($mencioA, $mencioB) = @_;
	my $spA = $mencioA->speaker;
	my $spB = $mencioB->speaker;
	if ($spA eq 'uknown' || $spB eq 'uknown') {
		return 2;
	}
	if ($mencioA->isEntity && $mencioA->isPerson &&  lc($mencioA->head_str) eq lc($spB)) {
		return 1;
	}
	if ($mencioB->isEntity && $mencioB->isPerson && lc($mencioB->head_str) eq lc($spA)) {
		return 1;
	}
	return 0;
}

1;
