package Mention;
require Markable;
our @ISA = qw(Markable);
use strict;

# Mention.pm preparat per Ontonotes (CONLL) en angles EN

###############################################################################
# globals

use Globals qw ($PATH_RESOURCES $WNLIB $LANG);
use lib $WNLIB;

# Semantic class
use WordNet::QueryData;

my %cacheSemClass;
my @SEMCLASSES = ('male#n#2', 'female#n#2', 'person#n#1', 'organization#n#3',
						  'location#n#1', 'date#n#1', 'time#n#2', 'money#n#1', 
						  'percent#n#1', 'object#n#1');

my $wn = undef;

my $NO_WORDNET = undef; # =1 No es pot fer servir el wordnet


###############################################################################
# functions

sub trueMention {
	my $m = shift;
	@_ ? $m->{_trueMention} = shift : $m->{_trueMention};
}

sub type
{
	my $m = shift;
	if (!defined($m->{_type})) {
		if ($m->isPronoun) {
			$m->{_type} = 'p';
		}
		elsif ($m->isEntity) {
			$m->{_type} = 'e';
		}
		else {
			$m->{_type} = 's';
		}
	}
	return $m->{_type};
}

sub newChain {
	my $m = shift;
	if (@_) {
		$m->{_newChain} = shift;
	}
	else {
		if (!defined($m->{_newChain})) {
			$m->{_newChain} = 0; #by default
		}
		return $m->{_newChain};
	}
}

sub sentence {
	my $m = shift;
	if (!defined($m->{_sentence})) {
		my $firstToken = $m->{span}[0];
		my $doc = $m->document;
		my $sentences = $doc->GetLevel('sentence');
		foreach my $sen (@$sentences) {
			if ($sen->{span_ini} <= $firstToken->index &&
				$sen->{span_fi} >= $firstToken->index) {
				$m->{_sentence} = $sen;
				last;
			}
		}
	}
	return $m->{_sentence};
}

sub inQuotes {
	my $m = shift;
	
	if (!defined($m->{_inQuotes})) {
		my $sentence = $m->sentence;
		my $dins = 0;
		my $doc = $m->document;
		my $pos = $doc->GetLevel('pos');
		foreach my $t (@{$sentence->{span}}) {
			last if ($t->{index} == $m->{span_ini});
			
			if ($pos->[$t->{pos_level}]->{punct} eq 'quotation') {
				$dins = 1 - $dins;
			}
			elsif ($t->{content} eq '``') {
				$dins = 1;
			}
			elsif ($t->{content} eq '\'\'') {
				$dins = 0;
			}
			elsif ($t->{content} eq '\"') {
				$dins = 1 - $dins;
			}
		}
		$m->{_inQuotes} = $dins;
	}
	return $m->{_inQuotes};
}

sub firstInSentence {
	my $m = shift;
	
	if (!defined($m->{_firstInSentence})) {
		my $sentence = $m->sentence;
		my $primer = 1;
		foreach my $t (@{$sentence->{span}}) {
			last if ($t->{index} >= $m->{span_ini});
			if (defined($t->{coref_level}) || defined($t->{mention_level})) {
				$primer = 0;
				last;
			}
		}
		$m->{_firstInSentence} = $primer;
	}
	return $m->{_firstInSentence};
}

sub isMaximalNP {
	my $m = shift;
	
	if (!defined($m->{_isMaximalNP})) {
		my $max = $m->GetMaximalNP;
		$m->{_isMaximalNP} = ($max->{id} eq $m->{id}) ? 1: 0;
	}
	return $m->{_isMaximalNP};
}

sub GetMaximalNP {
	my $m = shift;
	
	if (!defined($m->{_GetMaximalNP})) {
		my $last = $m;
		my $parent = $m->parentNP;
		while (defined($parent) && $parent->id ne $last->id 
			&& $parent->{span_text} ne $last->{span_text}) {
			$last = $parent;
			$parent = $last->parentNP;
		}
		$m->{_GetMaximalNP} = $last;
	}
	return $m->{_GetMaximalNP};
}

sub isPronoun
{
	my $m = shift;
	
	if (!defined($m->{_isPronoun})) {
		$m->{_isPronoun} = 0;
		my $doc = $m->document;
		my $pos = $doc->GetLevel('pos');
		if (scalar(@{$m->{span}}) == 1) {
			my $t = $m->{span}[0];
			my $tpos = $pos->[$t->{pos_level}]->{tag};
			if ($tpos =~ /^(PRP\$?|WP\$?)$/ || $tpos eq 'p') {
				$m->{_isPronoun} = 1;
			}
		}
	}
	return $m->{_isPronoun};
}

sub isPossessive
{
	my $m = shift;
	
	if (!defined($m->{_isPossessive})) {
		$m->{_isPossessive} = 0;
		my $doc = $m->document;
		my $pos = $doc->GetLevel('pos');
		if (scalar(@{$m->{span}}) == 1) {
			my $t = $m->{span}[0];
			my $tpos = $pos->[$t->{pos_level}]->{tag};
			if ($tpos =~ /^(PRP\$|WP\$|POS)$/) {
				$m->{_isPossessive} = 1;
			}
			else { # PoS extra
				$tpos = $pos->[$t->{pos_level}]->{postype};
				if ($tpos eq 'possessive') {
					$m->{_isPossessive} = 1;
				}
			}
		}
	}
	return $m->{_isPossessive};
}

# Hardcoded: English
sub isDefNP {
	my $m = shift;
	if (!defined($m->{_isDefNP})) {
		if ($LANG eq 'en') {
			$m->{_isDefNP} = ($m->{span}[0]->{content} =~ /^the$/i) ? 1 : 0;
		}
		else {
			$m->{_isDefNP} = 0; # default
			my $doc = $m->document;
			if ($doc->TestLevel('posextra')) {
				my $pos = $doc->GetLevel('pos');
				my $t = $m->{span}[0];
				my $tag = $pos->[$t->{pos_level}]->{tag};
				my $tpos = $pos->[$t->{pos_level}]->{postype};
				if ($tag eq 'd' && $tpos eq 'article') {
					$m->{_isDefNP} = 1;
				}
			}
		}
	}
	return $m->{_isDefNP};
}

# Hardcoded: English
sub isIndefNP {
	my $m = shift;
	if (!defined($m->{_isIndefNP})) {
		if ($LANG eq 'en') {
			$m->{_isIndefNP} = ($m->{span}[0]->{content} =~ /^(a|an|some|many|several|any|others?|anothers?|every|each)$/i) ? 1 : 0;
		}
		else {
			$m->{_isIndefNP} = 0; #default
			my $doc = $m->document;
			if ($doc->TestLevel('posextra')) {
				my $pos = $doc->GetLevel('pos');
				my $t = $m->{span}[0];
				my $tag = $pos->[$t->{pos_level}]->{tag};
				my $tpos = $pos->[$t->{pos_level}]->{postype};
				if ($tag eq 'd' && $tpos eq 'indefinite') {
					$m->{_isIndefNP} = 1;
				}
			}
		}
	}
	return $m->{_isIndefNP};
}

# Hardcoded: English
sub isDemNP {
	my $m = shift;
	if (!defined($m->{_isDemNP})) {
		if ($LANG eq 'en') {
			$m->{_isDemNP} = ($m->{span}[0]->{content} =~ /^(that|this|those|these)$/i) ? 1 : 0;
		}
		else {
			$m->{_isDemNP} = 0; #default
			my $doc = $m->document;
			if ($doc->TestLevel('posextra')) {
				my $pos = $doc->GetLevel('pos');
				my $t = $m->{span}[0];
				my $tag = $pos->[$t->{pos_level}]->{tag};
				my $tpos = $pos->[$t->{pos_level}]->{postype};
				if ($tag eq 'd' && $tpos eq 'demonstrative') {
					$m->{_isDemNP} = 1;
				}
			}
		}
	}
	return $m->{_isDemNP};
}

# Hardcoded: English
sub isPlural {
	my $m = shift;
	if (!defined($m->{_isPlural})) {
		if ($m->isComposite) {
			$m->{_isPlural} = 1;
		}
		else {
			my $doc = $m->document;
			my $pos = $doc->GetLevel('pos');
			my $head = $m->head;
			
			# 1-token mention. The head is a noun or proper name
			if (scalar(@{$m->{span}}) == 1 && !$m->isEntity) {
				if ($doc->TestLevel('posextra')) {
					my $num = $pos->[$m->{span}[0]->{pos_level}]->{num};
					if ($num eq 'p') {
						$m->{_isPlural} = 1;
					}
					elsif ($num eq 's') {
						$m->{_isPlural} = 0;
					}
				}
				else {
					my $p = $pos->[$m->{span}[0]->{pos_level}]->{tag};
					if ($p =~ /^(NNS|NNPS)$/) {
						$m->{_isPlural} = 1;
					}
					elsif ($p =~ /^(NN|NNP)$/) {
						$m->{_isPlural} = 0;
					}
					else {
						$head = $m->{span}[0];
					}
				}
			}
			# many tokens, we need the head
			else {
				$head = $m->head; # markable
				
				my $p;
				my $num;
				if ($m->isEntity) {
					my $t = $head->{span}[-1];
					$p = $pos->[$t->{pos_level}]->{tag};
					$num = $pos->[$t->{pos_level}]->{num};
				}
				else {
					$p = $pos->[$head->{pos_level}]->{tag};
					$num = $pos->[$head->{pos_level}]->{num};
				}
				
				if ($doc->TestLevel('posextra')) {
					if ($num eq 'p') {
						$m->{_isPlural} = 1;
					}
					elsif ($num eq 's') {
						$m->{_isPlural} = 0;
					}
				}
				else {
					if ($p =~ /^(NNS|NNPS)$/) {
						$m->{_isPlural} = 1;
					}
					elsif ($p =~ /^(NN|NNP)$/) {
						$m->{_isPlural} = 0;
					}
				}
			}
			
			if (!defined($m->{_isPlural})) {
				$m->{_isPlural} = 2; # default
				if ($LANG eq 'en') {
					my $content = $head->content;
					$content =~ s/self$// if ($m->isReflexive);
					if ($content =~ /^(I|me|he|she|its?|this|that|my|mine|his|him|hers?|one)$/i) {
						$m->{_isPlural} = 0;
					}
					elsif ($content =~ /^(we|us|they|them|theirs?|these|those|ours?)$/i) {
						$m->{_isPlural} = 1;
					}
					elsif ($content =~ /^(your?|yours)$/i) {
						$m->{_isPlural} = 2;
					}
				}
			}
		}
	}
	
	return $m->{_isPlural};
}

# Hardcoded: English
# Retorna 1ª 2ª o 3ª persona per pronoms (qualsevol altre cosa sempre es 3ª, no?)
sub person
{
	my $m = shift;
	
	if (!defined($m->{_person})) {
		$m->{_person} = 3; # 3rd person by default
		if ($m->isPronoun) {
			if ($LANG eq 'en') {
				my $str = $m->string;
				$str =~ s/self$// if ($m->isReflexive);
				if ($str =~ /^(I|me|my|mine|we|us|ours?)$/i) {
					$m->{_person} = 1;
				}
				elsif ($str =~ /^(your?|yours)$/i) {
					$m->{_person} = 2;
				}
			}
			else {
				my $doc = $m->document;
				if ($doc->TestLevel('posextra')) {
					my $pos = $doc->GetLevel('pos');
					if (defined($pos->[$m->head->index]->{person})) {
						$m->{_person} = $pos->[$m->head->index]->{person};
					}
				}
			}
		}
	}
	
	return $m->{_person};
}


sub isEliptic
{
	my $m = shift;
	if (!defined($m->{_isEliptic})) {
		$m->{_isEliptic} = 0;
		if ($m->isPronoun && $m->{span}[0]->content eq '_') {
			$m->{_isEliptic} = 1;
		}
	}
	return $m->{_isEliptic};
}


sub InitWN
{
	if (!defined($wn)) {
		$wn = WordNet::QueryData->new;
	}
}

# Mira si algun dels hyperonims (is-a) correspon a algun dels
# de la llista de classes @SEMCLASSES. Si cap correspon, seguim
# pujant per l'arbre cridant la funcio recursivament
sub ISA
{
	my $synset = shift;
	
	InitWN;
	my @hypes = $wn->querySense($synset, "hypes");
	if (@hypes) {
		foreach my $hype (@hypes) {
			my @c = grep( /^$hype$/, @SEMCLASSES);
			if (@c) {
				return $c[0];
			}
			else {
				my $class = ISA($hype);
				return $class if ($class ne 'unknown');
			}
		}
		return 'unknown';
	}
	else {
		return 'unknown';
	}
}

# Retorna classe semantica d'un synset
sub SemClass
{
	my $synset = shift;
	if (!defined($cacheSemClass{$synset})) {
		$cacheSemClass{$synset} = ISA($synset);
	}
	
	return $cacheSemClass{$synset};
}

# Hardcoded: English
# CoNLL (senses)
# lema
# Retorna Classe semantica d'una mencio
sub semanticClass
{
	my $m = shift;
	if (!defined($m->{_semanticClass})) {
		$m->{_semanticClass} = 'unknown'; #default
		my $generic = 1;
		
		if ($m->isPronoun) {
			$generic = 0;
			my $t = $m->{span}[0]->{content};
			$t =~ s/self$//i if ($m->isReflexive); # treiem el self dels reflexius
			if ($t =~ /^(he|his|him)$/i) {
				$m->{_semanticClass} = 'male#n#2';
			}
			elsif ($t =~ /^(she|hers?)$/i) {
				$m->{_semanticClass} = 'female#n#2';
			}
			elsif ($t =~ /^(i|you|us|me|we|my|mine|yours?|ours?)$/i) {
				$m->{_semanticClass} = 'person#n#1';
			}
			else { $generic = 1; }
		}
		elsif ($m->isEntity) {
			$generic = 0;
			my $ne = $m->GetEntity;
			if ($ne) {
				$ne->{tag} = lc($ne->{tag});
				$m->{_semanticClass} = 'person#n#1' if ($ne->{tag} eq 'person');
				$m->{_semanticClass} = 'organization#n#3' if ($ne->{tag} =~ /^org/);
				$m->{_semanticClass} = 'location#n#1' if ($ne->{tag} =~ /^loc/);
				$m->{_semanticClass} = 'location#n#1' if ($ne->{tag} eq 'gpe');
				$m->{_semanticClass} = 'location#n#1' if ($ne->{tag} eq 'fac');
				$m->{_semanticClass} = 'date#n#1' if ($ne->{tag} eq 'date');
				$m->{_semanticClass} = 'time#n#2' if ($ne->{tag} eq 'time');
				$m->{_semanticClass} = 'money#n#1' if ($ne->{tag} eq 'money');
				$m->{_semanticClass} = 'percent#n#1' if ($ne->{tag} eq 'percent');
				$m->{_semanticClass} = 'number#n#1' if ($ne->{tag} =~ /^(number|ordinal|cardinal|quantity)/);
				$m->{_semanticClass} = 'object#n#1' if ($ne->{tag} =~ /^(product|plant|animal|other)/);
			}
			else { $generic = 1; }
		}
		
		if ($generic) {
			if (defined($NO_WORDNET)) {
				$m->{_semanticClass} = 'unknown';
			}
			else {
				my $paraula = $m->head_str;
				
				my $sense = 1;
				if ($m->document->TestLevel('sense')) {
					my $sns = $m->document->GetLevel('sense');
					if (!$m->isEntity && $sns->[$m->head->index]->{tag} =~ /^\d+$/) {
						$sense = $sns->[$m->head->index]->{tag};
					}
				}
				
				my $synset = "$paraula#n#$sense";
				
				if ($m->document->TestLevel('lemma')) {
					my $lem = $m->document->GetLevel('lemma');
					if (!$m->isEntity) {
						my $lemma = $lem->[$m->head->index]->{tag};
						$synset = "$lemma#n#$sense";
					}
				}
				
				$m->{_semanticClass} = SemClass($synset);
			}
		}
	}
	
	return $m->{_semanticClass};
}


sub isPerson {
	my $m = shift;

	if (!defined($m->{_isPerson})) {
		$m->{_isPerson} = ($m->semanticClass =~ /(male|person)/) ? 1 : 0;
	}
	return $m->{_isPerson};
}

# Hardcoded: English
sub isReflexive {
	my $m = shift;
	if (!defined($m->{_isReflexive})) {
		$m->{_isReflexive} = 0; # default
		if ($LANG eq 'en') {
			$m->{_isReflexive} = ($m->isPronoun &&
									$m->{span}[0]->{content} =~ /sel(f|ves)$/i) ? 1: 0
		}
	}
	return $m->{_isReflexive};
}

# Hardcoded: English
sub gender
{
	my $m = shift;
	if (@_) {
		$m->{_gender} = shift;
		return;
	}
	return $m->{_gender} if (defined($m->{_gender}));
	
	if ($LANG eq 'en') {
		my $head_str = $m->head_str;
		if ($m->isPronoun) {
			if ($head_str =~ /^(he|him|himself|his)$/i) {
				$m->{_gender} = 'm';
			}
			elsif ($head_str =~ /^(she|her|herself|hers)$/i) {
				$m->{_gender} = 'f';
			}
			elsif ($head_str =~ /^(it|its|itself)/i) {
				$m->{_gender} = 'n';
			}
			else {
				$m->{_gender} = 'u';
			}
		}
		else {
			$m->{_gender} = 'u'; # default
			my $sem = $m->semanticClass;
			if ($sem =~ /^male/) {
				$m->{_gender} = 'm';
			}
			elsif ($sem =~ /^female/) {
				$m->{_gender} = 'f';
			}
			elsif ($sem =~ 
			/^(organization|time|date|object|location|money|percent)/) {
				$m->{_gender} = 'n';
			}
			elsif ($sem eq 'unknown') {
				$m->{_gender} = 'n'; # lo mes probable
			}
			elsif ($sem =~ /^person/) {
				# Mirem noms de persona
				my $head = $m->head;
				my $nom;
				if ($head->{mmax_level} eq 'enamex') {
					$nom = $head->{span}[0]->{content};
					# if social title is not useful for gender determination we
					# need the second word, hoping it is the name
					if ($nom =~ /^(Dr|Prof|Rev|Sen|Pres|MP|PC)\.?$/ && @{$head->{span}} > 1) {
						$nom = $head->{span}[1]->{content};
					}
				}
				else {
					$nom = $head->{content};
				}
				my $nom_net = $nom;
				$nom_net =~ s/[\'\"\(\)\&\$\[\]\`]//g;
				if ($nom_net =~ /^(Mr|Sir|Master|Mister|Esq|Lord)\.?$/) {
					$m->{_gender} = 'm';
				}
				elsif ($nom_net =~ /^(Mrs|Miss|Ms|Lady|Madam|Dame)\.?$/) {
					$m->{_gender} = 'f';
				}
				else {
					
					my ($female) = `grep -ic \'^$nom\$\' $PATH_RESOURCES/names-female.txt`;
					chomp($female);
					my ($male) = `grep -ic \'^$nom\$\' $PATH_RESOURCES/names-male.txt`;
					chomp($male);
					if ($female && !$male) {
						$m->{_gender} = 'f';
					}
					elsif (!$female && $male) {
						$m->{_gender} = 'm';
					}
					else {
						$m->{_gender} = 'u';
					}
				}
			}
		}
	}
	else {
		$m->{_gender} = 'u';
		my $head = $m->head;
		my $doc = $m->document;
		if ($doc->TestLevel('posextra')) {
			my $pos = $doc->GetLevel('pos');
			
			my $t;
			if ($m->isEntity) {
				$t = $head->{span}[0];
			}
			else {
				$t = $head;
			}
			
			my $gen = $pos->[$t->{pos_level}]->{gen};
			if ($gen =~ /^[mfn]$/) {
				$m->{_gender} = $gen;
			}
		}
	}
	
	return $m->{_gender};
}

sub isProperName {
	my $m = shift;
	if (!defined($m->{_isProperName})) {
		my $doc = $m->document;
		my $pos = $doc->GetLevel('pos');
		my $head = $m->head;
		my $mark;
		if ($m->isEntity) {
			$mark = $pos->[$head->{span_fi}];
		}
		else {
			$mark = $pos->[$head->{pos_level}];
		}
		
		if ($doc->TestLevel('posextra')) {
			my $tag = $mark->{tag};
			my $postype = $mark->{postype};
			$m->{_isProperName} = ($tag eq 'n' && $postype eq 'proper') ? 1 : 0;
		}
		else {
			$m->{_isProperName} = ($mark->{tag} =~ /^NNPS?$/) ? 1 : 0;
		}
	}
	return $m->{_isProperName};
}

# Torna 1 si es un nom comu, sino 0
sub isNoun
{
	my $m = shift;
	if (!defined($m->{_isNoun})) {
		my $doc = $m->document;
		my $pos = $doc->GetLevel('pos');
		my $head = $m->head;
		my $mark;
		if ($m->isEntity) {
			$mark = $pos->[$head->{span_fi}];
		}
		else {
			$mark = $pos->[$head->{pos_level}];
		}
		
		if ($doc->TestLevel('posextra')) {
			my $tag = $mark->{tag};
			my $postype = $mark->{postype};
			$m->{_isNoun} = ($tag eq 'n' && $postype eq 'common') ? 1 : 0;
		}
		else {
			$m->{_isNoun} = ($mark->{tag} =~ /^NNS?$/) ? 1 : 0;
		}
	}
	return $m->{_isNoun};
}


sub isEntity {
	my $m = shift;
	if (!defined($m->{_isEntity})) {
		my $doc = $m->document;
		my $head = $m->head;
		$m->{_isEntity} = ($head->{mmax_level} eq 'enamex') ? 1 : 0;
	}
	return $m->{_isEntity};
}

sub GetEntity {
	my $m = shift;
	return $m->isEntity ? $m->head : undef;
}

sub OrderMentions
{
	if ($a->{span_ini} < $b->{span_ini}) {
		return -1;
	}
	elsif ($a->{span_ini} > $b->{span_ini}) {
		return 1;
	}
	else {
		return (scalar(@{$b->{span}}) <=> scalar(@{$a->{span}}));
	}
}

# Hardcoded: English
# Michael Collins: Rules for NPs:
#  If the last word is tagged POS, return (last-word);
#  Else search from right to left for the first child which is an:
#     NN, NNP, NNPS, NNS, NX, POS, or JJR.
#  Else search from left to right for the first child which is an NP
#  Else search from right to left for the first child which is a $, ADJP or PRN.
#  Else search from right to left for the first child which is a CD.
#  Else search from right to left for the first child which is a JJ, JJS, RB or QP.
#  Else return the last word.
sub headSyn
{
	my $m = shift;
	my $head;
	my $doc = $m->document;
	my $tokens = $doc->tokens;
	my $syn = $doc->GetLevel('syn');
	my $pos = $doc->GetLevel('pos');
	if (defined($m->{span}) && @{$m->{span}} > 0) {
		$head = undef;
		if (@{$m->{span}} == 1) {
			$head = $m->{span}[0];
		}
		else {
			my @mask;
			my @fills;
			
			# Primer mirem si hi ha NPs fills
			for (my $j = 0; $j < @$syn; $j++) {
				my $npj = $syn->[$j];
				next if ($npj->{tag} ne 'NP');
				next if ($npj->{span_ini} == $m->{span_ini} && $npj->{span_fi} == $m->{span_fi});
				if ($npj->{span_ini} >= $m->{span_ini} && $npj->{span_fi} <= $m->{span_fi}) {
					push(@fills, $npj);
					for (my $k = 0; $k < @{$npj->{span}}; $k++) {
						$mask[$k + $npj->{span_ini} - $m->{span_ini}] = 1;
					}
				}
				if (@fills > 1) {
					@fills = sort OrderMentions (@fills);
				}
			}
			
			my $last = scalar(@{$m->{span}});
			if (@fills > 0) {
				while ($mask[$last] && $last >= 0) { $last--; }
			}
			
			if ($last >= 0 && $pos->[$m->{span_ini} + $last]->{tag} eq 'POS') {
				$head = $tokens->[$m->{span_ini} + $last];
			}
			else {
				foreach my $t (reverse(@{$m->{span}})) {
					if (!$mask[$t->index - $m->{span_ini}] &&
						$pos->[$t->index]->{tag} =~ /^(NNP?S?|NX|POS|JJR)$/) {
						$head = $t;
						last;
					}
				}
			}
		
			if (!defined($head)) {
				if (@fills) {
					$head = headSyn($fills[0]); # recursive!
				}
				else {
					foreach my $t (reverse(@{$m->{span}})) {
						if (!$mask[$t->index - $m->{span_ini}] &&
							$pos->[$t->index]->{tag} =~ /^(CD)$/) {
							$head = $t;
							last;
						}
					}
					
					if (!defined($head)) {
						foreach my $t (reverse(@{$m->{span}})) {
							if ($pos->[$t->index]->{tag} =~ /^(JJS?|RB|QP)$/) {
								$head = $t;
								last;
							}
						}
					}
				}
			}
			
			if (!defined($head)) {
				$head = $m->{span}[-1];
			}
		}
	}
	
	return $head;
}


# Retorna el nucli d'una mencio.
# Hardcoded: English
sub head
{
	my $m = shift;
	
	if (!defined($m->{_head})) {
		my $doc = $m->document;
		my $enamex = $doc->GetLevel('enamex');
		
		$m->{_head} = $m->{span}[-1]; # default
		
		# Try with constituent parsing
		if ($doc->TestLevel('syn') && $LANG eq 'en' && !$doc->TestLevel('posextra')) {
			$m->{_head} = $m->headSyn;
		}
		# Dependencies
		else {
		
			# En principi tots els tokens han de dependre d'un token que estigui dins de la mencio
			# excepte un que dependra d'un de fora o 0, que sera el nucli.
			my $dep = $doc->GetLevel('dep');
			my $sentence = $m->sentence;
			my $offset = $sentence->{span_ini} - 1;
			
			for (my $i = 0; $i < @{$m->{span}}; $i++) {
				my $token = $m->{span}[$i];
				my $apunta = $dep->[$token->{dep_level}]->{tag};
				if ($apunta == 0) {
					$m->{_head} = $token;
					last;
				}
				elsif ($apunta + $offset < $m->{span_ini} || $apunta + $offset > $m->{span_fi}) {
					$m->{_head} = $token;
					last;
				}
			}
		}
		

		if (defined($m->{_head}->{enamex_level})) {
			my $ne = $enamex->[$m->{_head}->{enamex_level}];
			# S'ha de comprovar que la NE no sigui mes llarga que la mencio!
			if ($m->{span_ini} <= $ne->{span_ini} && $ne->{span_fi} <= $m->{span_fi}) {
				$m->{_head} = $ne;
			}
		}

	}
	
	return $m->{_head};
}


sub head_str {
	my $m = shift;
	return $m->head->string;
}

# Hardcoded: English
# Les mencions que son una llista de mencions
#"the Department of Defense and the FBI" => head: Department of defense PERO son 2
#per tant es Composite
sub isComposite {
	my $m = shift;
	if (!defined($m->{_isComposite})) {
		
		my @fills = @{$m->{span}};
		if (scalar(@fills) == 1) {
			$m->{_isComposite} = 0;
		}
		else {
			$m->{_isComposite} = 0;
			my $comes = 0;
			my $doc = $m->document;
			my $pos = $doc->GetLevel('pos');
			foreach my $token (@{$m->{span}}) {
				if ($token->content =~ /^(and|or)$/ || $pos->[$token->index]->{tag} eq 'CC') {
					$m->{_isComposite} = 1;
					last;
				}
				elsif ($token->content =~ /^\,$/) {
					$comes++;
					if ($comes >= 2) {
						$m->{_isComposite} = 1;
						last;
					}
				}
				elsif ($pos->[$token->index]->{tag} eq 'c' &&
				$pos->[$token->index]->{postype} eq 'coordinating') {
					$m->{_isComposite} = 1;
					last;
				}
			}
		}
	}
	return $m->{_isComposite};
}

sub OrdenaParents
{
	if ($a->{span_ini} <= $b->{span_ini} &&
		$a->{span_fi} >= $b->{span_fi}) {
		return -1;
	}
	return 1;
}

sub parentNP {
	my $mencio = shift;
	if (!defined($mencio->{_parentNP_done})) {
		my $doc = $mencio->document;
		my $mencions = $doc->GetMentions($mencio->trueMention);
		my @parents;
		foreach my $m (@$mencions) {
			if ($m->{span_ini} <= $mencio->{span_ini} &&
				$m->{span_fi} >= $mencio->{span_fi} && $m->{id} ne $mencio->{id}) {
				# $m domina $mencio.
				push(@parents, $m);
			}
		}
		if (scalar(@parents) > 0) {
			@parents = sort OrdenaParents (@parents);
			$mencio->{_parentNP} = $parents[-1];
		}
		else {
			$mencio->{_parentNP} = undef;
		}
		$mencio->{_parentNP_done} = 1;
	}
	return $mencio->{_parentNP};
}


sub srl_npreds
{
	my $m = shift;
	if (!defined($m->{_srl_npreds})) {
		my $doc = $m->document;
		my $srl = $doc->GetLevel('srl');
		my $s = $srl->[$m->{span}[0]->{srl_level}];
		my $npreds = $s->{npreds};
		$m->{_srl_npreds} = $npreds;
	}
	return $m->{_srl_npreds};
}
 
 
sub ExtractApreds
{
	my ($token) = @_;
	return (0, []) if (!defined($token) || !UNIVERSAL::isa($token, 'Markable'));
	my $doc = $token->document;
	my $srl = $doc->GetLevel('srl');
	
	my $s = $srl->[$token->{srl_level}];
	my $npreds = $s->{npreds};
	my @apreds;
	my $trobat = 0;
	for (my $i = 0; $i < $npreds; $i++) {
		if (defined($s->{"apred$i"})) {
			$apreds[$i] = $s->{"apred$i"};
			$trobat = 1;
		}
		else {
			$apreds[$i] = undef;
		}
	}
	
	return ($trobat, \@apreds);
}

# Busca el primer tag de apreds a la mencio. si no en troba, fa el mateix pel parentNP
# Segueix fins que troba algo o no hi ha mes parents
sub srl_apreds
{
	my $m = shift;
	if (!defined($m->{_srl_apreds})) {
		my $doc = $m->document;
		my $srl = $doc->GetLevel('srl');
		
		my $trobat = 0;
		my $apreds;
		my $mention = $m;
		do {
			# Recorrem tots els tokens de la mencio
			foreach my $t (@{$mention->{span}}) {
				($trobat, $apreds) = ExtractApreds($t);
				if ($trobat) {
					$m->{_srl_apreds} = $apreds;
					last;
				}
			}
			$mention = $mention->parentNP;
		} while (defined($mention) && !$trobat);
		
		$m->{_srl_apreds} = $apreds;
	}
	return $m->{_srl_apreds};
}
 
# TODO: ara mateix es retorna l'apred que es trobi primer, pero s'hauria de
# veure en cas de tenir multiples apreds quin es el mes important
sub srl_arg
{
	my $m = shift;
	if (!defined($m->{_srl_arg})) {
		$m->{_srl_arg} = 'N'; # -1 significa cap
		my $doc = $m->document;
		my $srl = $doc->GetLevel('srl');
		my $apreds = $m->srl_apreds;
		for (my $i = 0; $i < @$apreds; $i++) {
			if (defined($apreds->[$i]) && $apreds->[$i] =~ /^(a|arg)/) {
				($m->{_srl_arg}) = ($apreds->[$i] =~ /^(?:a|arg)(\w)/);
				last;
			}
		}
	}
	return $m->{_srl_arg};
}

sub srl_role
{
	my $m = shift;
	if (!defined($m->{_srl_role})) {
		$m->{_srl_role} = -1; # -1 significa cap
		my $doc = $m->document;
		my $srl = $doc->GetLevel('srl');
		my $apreds = $m->srl_apreds;
		for (my $i = 0; $i < @$apreds; $i++) {
			if (defined($apreds->[$i]) && $apreds->[$i] =~ /^(a|arg)/) {
				($m->{_srl_role}) = ($apreds->[$i] =~ /^(?:a|arg).\-(\w+?)$/);
				last;
			}
		}
	}
	return $m->{_srl_role};
}


sub stringClean
{
	my $m = shift;
	
	if (!defined($m->{_string_clean})) {
		my $text;
		my $pos = $m->document->GetLevel('pos');
		
		# gets the content of each token
		if (defined($m->{span})) {
			foreach my $ref (@{$m->{span}}) {
				next if ($text eq '' && ($pos->[$ref->index]->{tag} !~ /^NNP?S?/
					|| $pos->[$ref->index]->{tag} eq 'n'));
				$text .= ' ' if ($text ne '');
				$text .= $ref->content;
			}
			$m->{_string_clean} = $text;
		}
		# tokens does not have span. We get the content
		else {
			$m->{_string_clean} = $m->content;
		}
		
		# substituim guions _ per espais
		$m->{_string_clean} =~ s/_/ /og;
	}
	
	return $m->{_string_clean};
}

sub vNouns
{
	my $m = shift;
	if (!defined($m->{_vNouns})) {
		my @vector;
		my $pos = $m->document->GetLevel('pos');
		
		# gets the content of each token
		if (defined($m->{span})) {
			foreach my $ref (@{$m->{span}}) {
				if ($pos->[$ref->index]->{tag} =~ /^NNP?S?/ ||
					$pos->[$ref->index]->{tag} eq 'n') {
					push(@vector, $ref->content);
				}
			}
			$m->{_vNouns} = \@vector;
		}
	}
	
	return $m->{_vNouns};
}

# Hardcoded: English
# Nouns a the inmediate left of the head (and the head itself)
sub term
{
	my ($m, $withoutHead) = @_;
	if (!defined($m->{_term}) || $withoutHead == 1) {
		my @vector;
		my $pos = $m->document->GetLevel('pos');
		my $head = $m->head;
		my $head_index = defined($head->{span_ini}) ? $head->{span_ini} : $head->{index};
		
		# gets the content of each token
		if (defined($m->{span})) {
			foreach my $ref (@{$m->{span}}) {
				last if ($ref->index >= $head_index);
				if ($pos->[$ref->index]->{tag} =~ /^NNP?S?/ ||
					$pos->[$ref->index]->{tag} =~ /^JJ/ ||
					$pos->[$ref->index]->{tag} =~ /^[an]$/) {
					push(@vector, $ref->content);
					}
					elsif ($ref->content eq "\'s") {
						# ho passem
					}
					else {
						@vector = ();
					}
			}
			my $str = join(' ', @vector);
			if (!$withoutHead) {
				$str .= ' ' if ($str ne '');
				$str .= $m->head_str;
				$m->{_term} = $str;
			}
			else {
				return $str;
			}
		}
	}
	
	return $m->{_term};
}

# sub vNouns_prev
# {
# 	my $m = shift;
# 	if (!defined($m->{_vNouns_prev})) {
# 		my @vector;
# 		my $pos = $m->document->GetLevel('pos');
# 		my $head = $m->head;
# 		my $head_index = defined($head->{span_ini}) ? $head->{span_ini} : $head->{index};
# 		
# 		# gets the content of each token
# 		if (defined($m->{span})) {
# 			foreach my $ref (@{$m->{span}}) {
# 				last if ($ref->index >= $head_index);
# 				if ($pos->[$ref->index]->{tag} =~ /^NNP?S?/) {
# 					push(@vector, $ref->content);
# 				}
# 			}
# 			$m->{_vNouns_prev} = \@vector;
# 		}
# 	}
# 	
# 	return $m->{_vNouns_prev};
# }
# 
# sub vNouns_post
# {
# 	my $m = shift;
# 	if (!defined($m->{_vNouns_post})) {
# 		my @vector;
# 		my $pos = $m->document->GetLevel('pos');
# 		my $head = $m->head;
# 		my $head_index = defined($head->{span_ini}) ? $head->{span_ini} : $head->{index};
# 		
# 		# gets the content of each token
# 		if (defined($m->{span})) {
# 			foreach my $ref (@{$m->{span}}) {
# 				if ($ref->index > $head_index) {
# 					if ($pos->[$ref->index]->{tag} =~ /^NNP?S?/) {
# 						push(@vector, $ref->content);
# 					}
# 				}
# 			}
# 			$m->{_vNouns_post} = \@vector;
# 		}
# 	}
# 	
# 	return $m->{_vNouns_post};
# }

sub vAdjectives
{
	my $m = shift;
	if (!defined($m->{_vAdjectives})) {
		my @vector;
		my $pos = $m->document->GetLevel('pos');
		
		# gets the content of each token
		if (defined($m->{span})) {
			foreach my $ref (@{$m->{span}}) {
				if ($pos->[$ref->index]->{tag} =~ /^JJ/ ||
					$pos->[$ref->index]->{tag} eq 'a') {
					push(@vector, $ref->content);
				}
			}
			$m->{_vAdjectives} = \@vector;
		}
	}
	
	return $m->{_vAdjectives};
}

sub speaker
{
	my $m = shift;
	
	if (!defined($m->{_speaker})) {
		my $doc = $m->document;
		if ($doc->TestLevel('speaker')) {
			my $spk = $doc->GetLevel('speaker');
			$m->{_speaker} = $spk->[$m->index]->{tag};
		}
		else {
			$m->{_speaker} = 'uknown';
		}
	}
	return $m->{_speaker};
}

1;
