package Model;
use strict;
use Document;
use Features;
use Globals qw ($PATH_RESOURCES);

sub new {
	my ($pkg, $file) = @_;
	my $model = bless { _modelFile => $file }, $pkg;
	$model->modelFile($file);
	$model->trueMentions(0); # by default
	$model->verbosity(1); # by default
	
	return $model;
}

sub verbosity {
	my $obj = shift;
	@_ ? $obj->{_verbosity} = shift : $obj->{_verbosity};
}

sub modelFile {
	my $obj = shift;
	if (@_) {
		my $file = shift;
		$obj->{_modelFile} = $file;
		
		if ($file =~ /^(.*)\/([^\/]*)$/) {
			my $dir = $1;
			my $name = $2;
			if (!-e $dir) {
				mkdir($dir) || die "Error ($dir): $!";
			}
		}
		
		$obj->datasetFile("$file.data");
	}
	else {
		return $obj->{_modelFile};
	}
}

sub datasetFile {
	my $obj = shift;
	@_ ? $obj->{_datasetFile} = shift : $obj->{_datasetFile};
}

sub mask {
	my $obj = shift;
	if (@_) {
		$obj->{_mask} = shift;
		$obj->{_features} = undef;
		$obj->{_invMask} = undef;
	}
	else {
		return $obj->{_mask};
	}
}

sub inverseMask {
	my $obj = shift;
	if (@_) {
		$obj->{_invMask} = shift;
		$obj->{_mask} = undef;
		$obj->{_features} = undef;
		$obj->{_mask} = $obj->features;
		foreach my $key (keys(%{$obj->{_invMask}})) {
			$obj->{_mask}->{$key} = undef;
		}
		$obj->{_features} = undef;
	}
	else {
		return $obj->{_invMask};
	}
}

sub alias {
	my $obj = shift;
	@_ ? $obj->{_alias} = shift : $obj->{_alias};
}

sub trueMentions {
	my $obj = shift;
	@_ ? $obj->{_trueMentions} = shift : $obj->{_trueMentions};
}

sub features {
	my $obj = shift;
	if (!defined($obj->{_features})) {
		my @features;
		my %iFeat;
		my $i = 0;
		open (L, "$PATH_RESOURCES/features.txt") || die "ooh";
		while (my $l=<L>) {
			chomp($l);
			my ($a) = split(/\:/, $l);
			unless (defined($obj->{_mask}) && defined($obj->{_mask}->{$a})) {
				$features[$i] = $a;
				$iFeat{$a} = $i;
				$i++;
			}
		}
		close (L);
		$obj->{_features} = \@features;
		$obj->{_iFeatures} = \%iFeat;
	}
	
	if (wantarray) {
		return @{$obj->{_features}};
	}
	return $obj->{_iFeatures};
}


sub groupFeatures {
	my ($obj, $N) = @_;
	my @gfeat;
	my $k=0;
	my %igfeat;
	if (!defined($obj->{'_groupFeatures' . $N})) {
		for (my $i = 0; $i < $N - 1; $i++) {
			for (my $j = $i+1; $j < $N; $j++) {
				foreach my $key (keys(%{$obj->features})) {
					my $str = $key . '_' . $i . $j;
					if ($key =~ /^I_(.*?)$/) {
						$str = $1 . '_' . $i;
					}
					elsif ($key =~ /^J_(.*?)$/) {
						$str = $1 . '_' . $j;
					}
					
					
# 					if (!defined($igfeat{$str})) {
						$gfeat[$k] = $str;
						$igfeat{$str} = $k;
						$k++;
# 					}
					
				}
			}
		}
		$obj->{'_groupFeatures' . $N} = \@gfeat;
		$obj->{'_iGroupFeatures' . $N} = \%igfeat;
	}
	
	if (wantarray) {
		return @{$obj->{'_groupFeatures' . $N}};
	}
	return $obj->{'_iGroupFeatures' . $N};
}

sub GenPairFeatures {
	my ($obj, $m_i, $m_j) = @_;
	my %a;
	
	$obj->GenStructural(\%a, $m_i, $m_j);
	$obj->GenLexical(\%a, $m_i, $m_j);
	$obj->GenMorpho(\%a, $m_i, $m_j);
	$obj->GenSyntactic(\%a, $m_i, $m_j);
	$obj->GenSemantic(\%a, $m_i, $m_j);
	$obj->GenDiscourse(\%a, $m_i, $m_j);

	if (wantarray) {
		my @ford = $obj->features;
		my @features;
		for (my $i = 0; $i < @ford; $i++) {
			$features[$i] = $a{$ford[$i]};
		}
		return @features;
	}

	return \%a;
}

sub GenStructural
{
	my ($obj, $a, $mencioA, $mencioB) = @_;
	
	# structural
	my $dists = &Features::DistSentences( $mencioA, $mencioB, 0);
	$a->{DIST_SEN_0} = ($dists == 0) ? 1 : 0;
	$a->{DIST_SEN_1} = ($dists == 1) ? 1 : 0;
	$a->{DIST_SEN_L3} = ($dists <= 3) ? 1 : 0;
	$a->{DIST_SEN_G3} = ($dists > 3) ? 1 : 0;
	
	my $distm = &Features::DistMentions($mencioA, $mencioB, 0) - 1;
	$a->{DIST_MEN_0} = ($distm == 0) ? 1 : 0;
	$a->{DIST_MEN_L3} = ($distm <= 3) ? 1 : 0;
	$a->{DIST_MEN_L10} = ($distm <= 10) ? 1 : 0;
	$a->{DIST_MEN_G10} = ($distm > 10) ? 1 : 0;
	
	my $distp = &Features::DistPhrases( $mencioA, $mencioB);
	$a->{DIST_PHR_0} = ($distp == 0) ? 1 : 0;
	$a->{DIST_PHR_1} = ($distp == 1) ? 1 : 0;
	$a->{DIST_PHR_L3} = ($distp <= 3 && $distp >= 0) ? 1 : 0;
	
	$a->{I_IN_QUOTES} = &Features::InQuotes($mencioA);
	$a->{J_IN_QUOTES} = &Features::InQuotes($mencioB);
	$a->{I_FIRST} = &Features::FirstInSentence($mencioA);
	$a->{J_FIRST} = &Features::FirstInSentence($mencioB);
	$a->{APPOSITIVE} = &Features::Appositive($mencioA, $mencioB);
	$a->{NESTED} = &Features::Nested($mencioA, $mencioB);
	
	my $ti = &Features::MentionType($mencioA);
	$a->{I_TYPE_P} = ($ti eq 'p') ? 1 : 0;
	$a->{I_TYPE_S} = ($ti eq 's') ? 1 : 0;
	$a->{I_TYPE_E} = ($ti eq 'e') ? 1 : 0;
	
	my $tj = &Features::MentionType($mencioB);
	$a->{J_TYPE_P} = ($tj eq 'p') ? 1 : 0;
	$a->{J_TYPE_S} = ($tj eq 's') ? 1 : 0;
	$a->{J_TYPE_E} = ($tj eq 'e') ? 1 : 0;
	
}

sub GenLexical
{
	my ($obj, $a, $mencioA, $mencioB) = @_;
	
	# lexical
	$a->{STR_MATCH} = &Features::StringMatch($mencioA, $mencioB);
	$a->{PRO_STR} = &Features::PRO_STR($mencioA, $mencioB);
	$a->{PN_STR} = &Features::PN_STR($mencioA, $mencioB);
	$a->{SOON_STR_NONPRO} = &Features::SOON_STR_NONPRO($mencioA, $mencioB);
	$a->{I_HEAD_TERM} = &Features::SameHeadTerm($mencioA);
	$a->{J_HEAD_TERM} = &Features::SameHeadTerm($mencioB);
	$a->{HEAD_MATCH} = &Features::HeadMatch($mencioA, $mencioB);
	$a->{TERM_MATCH} = &Features::TermMatch($mencioA, $mencioB);
	my $alias = &Features::Alias($mencioA, $mencioB);
	$a->{ALIAS_YES} = ($alias == 1) ? 1 : 0;
	$a->{ALIAS_NO} = ($alias == 0) ? 1 : 0;
	$a->{ALIAS_UN} = ($alias == 2) ? 1 : 0;
}

sub GenMorpho
{
	my ($obj, $a, $mencioA, $mencioB) = @_;
	
	# Morphological
	$a->{I_POSSESSIVE} = &Features::IsPossessive($mencioA);
	$a->{J_POSSESSIVE} = &Features::IsPossessive($mencioB);
	
	my $number = &Features::NumberMatch($mencioA, $mencioB);
	$a->{NUMBER_YES} = ($number == 1) ? 1 : 0;
	$a->{NUMBER_NO} = ($number == 0) ? 1 : 0;
	$a->{NUMBER_UN} = ($number == 2) ? 1 : 0;
	
	my $gender = &Features::GenderMatch($mencioA, $mencioB);
	$a->{GENDER_YES} = ($gender == 1) ? 1 : 0;
	$a->{GENDER_NO} = ($gender == 0) ? 1 : 0;
	$a->{GENDER_UN} = ($gender == 2) ? 1 : 0;
	
	$a->{I_THIRD_PERSON} = &Features::Is3Person($mencioA);
	$a->{J_THIRD_PERSON} = &Features::Is3Person($mencioB);
	
	$a->{I_PROPER_NAME} = &Features::ProperName($mencioA);
	$a->{J_PROPER_NAME} = &Features::ProperName($mencioB);
	
	$a->{I_NOUN} = &Features::IsNoun($mencioA);
	$a->{J_NOUN} = &Features::IsNoun($mencioB);
	
	my $agr = &Features::AGREEMENT($mencioA, $mencioB);
	$a->{AGREEMENT_YES} = ($agr == 1) ? 1 : 0;
	$a->{AGREEMENT_NO} = ($agr == 0) ? 1 : 0;
	$a->{AGREEMENT_UN} = ($agr == 2) ? 1 : 0;
	
	my $cagr = &Features::ClosestAgreement($mencioA, $mencioB);
	$a->{C_AGREEMENT_YES} = ($cagr == 1) ? 1 : 0;
	$a->{C_AGREEMENT_NO} = ($cagr == 0) ? 1 : 0;
	$a->{C_AGREEMENT_UN} = ($cagr == 2) ? 1 : 0;
	
	$a->{I_REFLEXIVE} = &Features::IsReflexive($mencioA);
	$a->{J_REFLEXIVE} = &Features::IsReflexive($mencioB);
}

sub GenSyntactic
{
	my ($obj, $a, $mencioA, $mencioB) = @_;
	
	# Syntactic
	$a->{I_DEF_NP} = &Features::DefNP($mencioA);
	$a->{J_DEF_NP} = &Features::DefNP($mencioB);
	$a->{I_DEM_NP} = &Features::DemNP($mencioA);
	$a->{J_DEM_NP} = &Features::DemNP($mencioB);
	$a->{MAXIMALNP} = &Features::MaximalNP($mencioA, $mencioB);
	$a->{I_MAXIMALNP} = &Features::IsMaximalNP($mencioA);
	$a->{J_MAXIMALNP} = &Features::IsMaximalNP($mencioB);
	$a->{I_INDEF_NP} = &Features::IndefNP($mencioA);
	$a->{J_INDEF_NP} = &Features::IndefNP($mencioB);
	$a->{I_EMBEDDED} = &Features::IsEmbeddedNoun($mencioA);
	$a->{J_EMBEDDED} = &Features::IsEmbeddedNoun($mencioB);
	$a->{I_COMPOSITE} = &Features::IsComposite($mencioA);
	$a->{J_COMPOSITE} = &Features::IsComposite($mencioB);
	$a->{BINDING_POS} = &Features::BindingPos($mencioA, $mencioB);
	$a->{BINDING_NEG} = &Features::BindingNeg($mencioA, $mencioB);
	$a->{C_COMMANDS_IJ} = &Features::C_Commands($mencioA, $mencioB);
	$a->{C_COMMANDS_JI} = &Features::C_Commands($mencioB, $mencioA);

	# Semeval
	$a->{SRL_SAMEVERB} = &Features::SRL_SameVerb( $mencioA, $mencioB);
	my $iarg = &Features::SRL_Arg( $mencioA);
	$a->{I_SRL_ARG_0} = ($iarg =~ /^\d$/ && $iarg == 0) ? 1 : 0;
	$a->{I_SRL_ARG_1} = ($iarg =~ /^\d$/ && $iarg == 1) ? 1 : 0;
	$a->{I_SRL_ARG_2} = ($iarg =~ /^\d$/ && $iarg == 2) ? 1 : 0;
	$a->{I_SRL_ARG_X} = ($iarg =~ /^\d$/ && $iarg > 2) ? 1 : 0;
	$a->{I_SRL_ARG_NUM} = ($iarg =~ /^\d$/) ? 1 : 0;
	$a->{I_SRL_ARG_N} = ($iarg eq 'N') ? 1 : 0;
	$a->{I_SRL_ARG_M} = ($iarg eq 'M') ? 1 : 0;
	$a->{I_SRL_ARG_L} = ($iarg eq 'L') ? 1 : 0;
	$a->{I_SRL_ARG_Z} = ($iarg !~ /^[\dNML]$/) ? 1 : 0;
	
	my $jarg = &Features::SRL_Arg( $mencioB);
	$a->{J_SRL_ARG_0} = ($jarg =~ /^\d$/ && $jarg == 0) ? 1 : 0;
	$a->{J_SRL_ARG_1} = ($jarg =~ /^\d$/ && $jarg == 1) ? 1 : 0;
	$a->{J_SRL_ARG_2} = ($jarg =~ /^\d$/ && $jarg == 2) ? 1 : 0;
	$a->{J_SRL_ARG_X} = ($jarg =~ /^\d$/ && $jarg > 2) ? 1 : 0;
	$a->{J_SRL_ARG_NUM} = ($jarg =~ /^\d$/) ? 1 : 0;
	$a->{J_SRL_ARG_N} = ($jarg eq 'N') ? 1 : 0;
	$a->{J_SRL_ARG_M} = ($jarg eq 'M') ? 1 : 0;
	$a->{J_SRL_ARG_L} = ($jarg eq 'L') ? 1 : 0;
	$a->{J_SRL_ARG_Z} = ($jarg !~ /^[\dNML]$/) ? 1 : 0;
	
	$a->{SAME_SRL_ARG} = &Features::SRL_SameArg( $mencioA, $mencioB);
}

sub GenSemantic
{
	my ($obj, $a, $mencioA, $mencioB) = @_;
	
	# Semantic
	$a->{VERB_IS} = &Features::VerbIs($mencioA, $mencioB);
	my $sem = &Features::SemanticClassMatch($mencioA, $mencioB);
	$a->{SEMCLASS_YES} = ($sem == 1) ? 1 : 0;
	$a->{SEMCLASS_NO} = ($sem == 0) ? 1 : 0;
	$a->{SEMCLASS_UN} = ($sem == 2) ? 1 : 0;
	
	$a->{I_PERSON} = &Features::IsPerson($mencioA);
	$a->{J_PERSON} = &Features::IsPerson($mencioB);
	$a->{I_ORGANIZATION} = &Features::IsOrganization($mencioA);
	$a->{J_ORGANIZATION} = &Features::IsOrganization($mencioB);
	$a->{I_LOCATION} = &Features::IsLocation($mencioA);
	$a->{J_LOCATION} = &Features::IsLocation($mencioB);
	
	$a->{ANIMACY} = &Features::Animacy($mencioA, $mencioB);
	$a->{INCOMPATIBLES} = &Features::Incompatibles( $mencioA, $mencioB);
	
	$a->{SRL_SAME_ROLE} = &Features::SRL_SameRole( $mencioA, $mencioB);
}

#         1 = revision
#         2 = cause
#         3 = equality
#         4 = context
sub GenDiscourse
{
	my ($obj, $a, $mencioA, $mencioB) = @_;
	
	my $sameSpeaker = &Features::SameSpeaker($mencioA, $mencioB);
	$a->{SAME_SPEAKER_YES} = ($sameSpeaker == 1) ? 1 : 0;
	$a->{SAME_SPEAKER_NO} = ($sameSpeaker == 0) ? 1 : 0;
	$a->{SAME_SPEAKER_UN} = ($sameSpeaker == 2) ? 1 : 0;
	
	my $sMatch = &Features::SpeakerMatch($mencioA, $mencioB);
	$a->{SPEAKER_MATCH_YES} = ($sMatch == 1) ? 1 : 0;
	$a->{SPEAKER_MATCH_NO} = ($sMatch == 0) ? 1 : 0;
	$a->{SPEAKER_MATCH_UN} = ($sMatch == 2) ? 1 : 0;
	
	my $dsMark = &Features::DiscourseLinked($mencioA, $mencioB);
	$a->{DS_MARK_NO} = ($dsMark == 0) ? 1 : 0;
	$a->{DS_MARK_RE} = ($dsMark == 1) ? 1 : 0;
	$a->{DS_MARK_CA} = ($dsMark == 2) ? 1 : 0;
	$a->{DS_MARK_EQ} = ($dsMark == 3) ? 1 : 0;
	$a->{DS_MARK_CO} = ($dsMark == 4) ? 1 : 0;
	
	$dsMark = &Features::HasDiscourseMarker( $mencioA);
	$a->{I_DS_MARK_NO} = ($dsMark == 0) ? 1 : 0;
	$a->{I_DS_MARK_RE} = ($dsMark == 1) ? 1 : 0;
	$a->{I_DS_MARK_CA} = ($dsMark == 2) ? 1 : 0;
	$a->{I_DS_MARK_EQ} = ($dsMark == 3) ? 1 : 0;
	$a->{I_DS_MARK_CO} = ($dsMark == 4) ? 1 : 0;
	
	$dsMark = &Features::HasDiscourseMarker( $mencioB);
	$a->{J_DS_MARK_NO} = ($dsMark == 0) ? 1 : 0;
	$a->{J_DS_MARK_RE} = ($dsMark == 1) ? 1 : 0;
	$a->{J_DS_MARK_CA} = ($dsMark == 2) ? 1 : 0;
	$a->{J_DS_MARK_EQ} = ($dsMark == 3) ? 1 : 0;
	$a->{J_DS_MARK_CO} = ($dsMark == 4) ? 1 : 0;
}

# Genera atributs per N mencions (tots amb tots)
sub GenGroupFeatures
{
	my ($obj, @mencions) = @_;
	my %atr;
	my %a;
	for (my $i = 0; $i < @mencions - 1; $i++) {
		for (my $j = $i+1; $j < @mencions; $j++) {
			my $pair = "$i,$j";
			$atr{$pair} = $obj->GenPairFeaturesExt($mencions[$i], $mencions[$j]);
			foreach my $key (keys(%{$atr{$pair}})) {
				my $str = $key . '_' . $i . $j;
				if ($key =~ /^I_(.*?)$/) {
					$str = $1 . '_' . $i;
				}
				elsif ($key =~ /^J_(.*?)$/) {
					$str = $1 . '_' . $j;
				}
				
				$a{$str} = $atr{$pair}->{$key};
			}
		}
	}
	return \%a;
}


sub constraints
{
	my ($obj, $c) = @_;
	if (!defined($c)) {
		if (!defined($obj->{_constraints})) {
			my @constraints = $obj->LoadConstraints;
			$obj->{_constraints} = \@constraints;
		}
		return $obj->{_constraints};
	}
	else {
		$obj->SaveConstraints($c);
		$obj->{_constraints} = $c;
	}
}


# Guarda les regles en un fitxer de text
sub SaveConstraints
{
	my ($obj, $c) = @_;
	my $fBase = $obj->modelFile;
	my $file = $fBase . '.constraints';
	
	if (open (R, ">$file")) {
		print "Saving " . scalar(@$c) . " constraints.\n" if ($obj->verbosity);
		for (my $k = 0; $k < scalar(@$c); $k++) {
			my ($cond, $val) = @{$c->[$k]};
			print R $cond . '#' , $val . "\n";
		}
		close (R);
	}
	else {
		print "ERROR: Can not save the constraints ($file) $!\n";
	}
}

# Carrega les regles d'un fitxer de text (format que fa servir SaveRules)
sub LoadConstraints
{
	my ($obj) = @_;
	my $fBase = $obj->modelFile;
	my $file = $fBase . '.constraints';
	my @c;
	
	open (R, $file) || die "Can not open $file: $!";
	my $firstline = 1;
	my $versio = 1;
	while (my $l = <R>) {
		chomp($l);
		next if ($l eq '');
		if ($firstline == 1 && $l =~ /^;VERSION (\d+)$/) {
			$versio = $1;
			print "Constraints version: $versio\n" if ($obj->verbosity);
			$firstline = 0;
			next;
		}
		$firstline = 0;
		if ($versio == 1) {
			my ($cond, $val) = split(/\#/, $l);
			push(@c, [$cond, $val]);
		}
		elsif ($versio == 2) {
			# ignore lines starting by ;
			if ($l =~ /^;/) {
				next;
			}
			my %c2;
			my ($reb, $inf, $cond) = split(/\#/, $l);
			$cond = '$r = ' . $cond if ($cond !~ /^\$r /);
			$c2{cond} = $cond;
			$c2{reb} = $reb;
			my $ii = 0;
			foreach my $rr (split(/\:/, $inf)) {
				@{$c2{inf}[$ii]} = split(/\,/, $rr);
				$ii++;
			}
			push(@c, [$cond, \%c2]);
		}
	}
	close (R);
	print scalar(@c) . " constraints\n" if ($obj->verbosity);
	return @c;
}


sub weights
{
	my ($obj, $w) = @_;
	if (!defined($w)) {
		if (!defined($obj->{_weights})) {
			my @weights = $obj->LoadWeights;
			$obj->{_weights} = \@weights;
		}
		return $obj->{_weights};
	}
	else {
		$obj->SaveWeights($w);
		$obj->{_weights} = $w;
	}
}

sub SaveWeights
{
	my ($obj, $w) = @_;
	my $wfile = $obj->modelFile . '.weights';
	open (PESOS, ">$wfile") || die "Can not write $wfile";
	print PESOS join(',', @$w);
	close (PESOS);
}

sub LoadWeights
{
	my ($obj) = @_;
	my $wfile = $obj->modelFile . '.weights';
	open (PESOS, $wfile) || die "Can not read $wfile";
	my $l = <PESOS>;
	chomp($l);
	my @w = split(/\,/, $l);
	close (PESOS);
	return @w;
}

1;
