package RelaxCor;

# RelaxCor
# Copyright (C) 2011, Emili Sapena esapena <at> lsi.upc.edu
# 
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version. This program is distributed in the hope that
# it will be useful, but WITHOUT ANY WARRANTY; without even the implied
# warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

use MentionDetection;
use Document;
use DT;
use SolverRelax;
use strict;

use Math::Round;

###############################################################################
# globals

our $VERSION = '1.0';


###############################################################################
# functions

sub LoadConfig
{
	my ($file) = @_;
	my %config;
	my $fixcol = 0;
	open (C, $file) || die "Can not open config file $file";
	while (my $l=<C>) {
		chomp($l);
		$l =~ s/\#.*$//;
		next if ($l !~ /\,/);
		my ($k, $v) = split(/\,/, $l);
		$k =~ s/^\s+//og;
		$k =~ s/\s+$//og;
		$v =~ s/^\s+//og;
		$v =~ s/\s+$//og;
		$config{$k} = $v;
	}
	
	return \%config;
}


sub GenData
{
	my ($model, $doc, $train) = @_;
	
	my $mentions = $doc->GetMentions($model->trueMentions, $train);
	my $file = $doc->file;
	
	# Save the number of mentions
	open (OUT, ">$file.nm") || die "No puc obrir $file.nm";
	print OUT scalar(@$mentions);
	close (OUT);
	
	# Generate an array of features for each possible pair of mentions in the doc
	open (OUT, ">$file.data") || die "No puc obrir $file.data";
	
	# first line with feature names
	print OUT join(',', $model->features) . "\n";
	
	for (my $i = 0; $i < @$mentions - 1; $i++) {
		for (my $j = $i+1; $j < @$mentions; $j++) {
			my @atributs = $model->GenPairFeatures($mentions->[$i], $mentions->[$j]);
			print OUT join(',', @atributs);
			if ($train) {
				my $result = $doc->Corefer($mentions->[$i], $mentions->[$j]) ? '+' : '-';
				print OUT ",$result";
			}
			print OUT "\n";
		}
	}
	close (OUT);
}

sub LoadData
{
	my ($model, $file) = @_;
	
	if (open (DAT, "$file.data")) {
		print "Loading...\n";
		my @tot;
		
		my $l = <DAT>;
		chomp($l);
		my @feat = split(/\,/, $l);
		
		# Check that the model features are included in the data
		my %datafeat;
		foreach my $f (@feat) {
			$datafeat{$f} = 1;
		}
		my $ifeat = $model->features;
		foreach my $key (keys(%{$ifeat})) {
			if (!defined($datafeat{$key})) {
				die "Features in cache data do not fit with model features ($key)";
			}
		}
		
		while ($l=<DAT>) {
			chomp ($l);
			my @attr = split(/\,/, $l);
			pop(@attr) if ($attr[-1] =~ /[\+\-]/); #treiem classe
			
			my @ordfeat;
			for (my $i = 0; $i < @feat; $i++) {
				$ordfeat[$ifeat->{$feat[$i]}] = $attr[$i] if (defined($ifeat->{$feat[$i]}));
			}
			push(@tot, \@ordfeat);
		}
		close (DAT);
		print "OK\n";
		return \@tot;
	}

	return undef;
}

# Check compatibility of the features in the file with the features expected by the model
sub CheckFeatures
{
	my ($model, $file, $attOrd, $N) = @_;
	my @features = $model->features;
# 	my @featuresN;
	my $needsReorder = 0;
	if (defined($N) && $N > 2) {
		@features = $model->groupFeatures($N);
		$needsReorder = 1;
	}
	%{$attOrd} = ();
	
	
	open (F, "$file.data") || die "oops $file.data: $!";
	
	# first line has the names and order of the features in the .data file
	my $attrline = <F>;
	chomp($attrline);
	my @order = split(/\,/, $attrline);
	
	# mention-pair
	if (!defined($N) || $N == 2) {
		for (my $i = 0; $i < @order; $i++) {
			$attOrd->{$order[$i]} = $i;
		}
	}
	# mention-group or entity-mention
	else {
		my $index = 0;
		for (my $i = 0; $i < $N; $i++) {
			for (my $j = $i+1; $j < $N; $j++) {
				for (my $k = 0; $k < @order; $k++) {
					my $str = $order[$k] . '_' . $i . $j;
					if ($order[$k] =~ /^I_(.*?)$/) {
						$str = $1 . '_' . $i;
					}
					elsif ($order[$k] =~ /^J_(.*?)$/) {
						$str = $1 . '_' . $j;
					}
					$attOrd->{$str} = $index;
# 					print "$str => $index\n";
					$index++;
				}
			}
		}
	}
	
	# Check compatibility of the features in the file with the features expected by the model
	for (my $i = 0; $i < @features; $i++) {
		if (!defined($attOrd->{$features[$i]})) {
			die "Features in data file are incompatible with the features expected by the model ($features[$i])";
		}
	}
	
	# The order of the columns in the data file can be different than
	# the order of the features expected by the model
	if (scalar(@order) != scalar(@features)) {
		$needsReorder = 1;
	}
	else {
		for (my $i = 0; $i < @features; $i++) {
			if ($attOrd->{$features[$i]} != $i) {
				$needsReorder = 1;
				last;
			}
		}
	}
	close (F);
	
	return $needsReorder;
}

# Gets the data examples of each file that agree with the conditions and
# writes them to the corresponding dataset files
sub SplitData
{
	my ($model, $trainFiles, $condI, $condJ) = @_;
	my $outfile = $model->datasetFile;
	my @features = $model->features;
	
	open (POS, ">$outfile.pos") || die "ooh $!";
	open (NEG, ">$outfile.negtots") || die "ooh $!";
	
	# agafem tots els exemples positius i negatius de cada fitxer
	foreach my $file (@$trainFiles) {
		my ($pos, $neg) = (0, 0);
		my %attOrd;
		my $needsReorder = CheckFeatures($model, $file, \%attOrd);
		open (F, "$file.data") || die "oops $file.data: $!";
		
		# first line
		my $attrline = <F>;
		
		while (my $l = <F>) {
			chomp($l);
			my @attr = split(/\,/, $l);
			if ( (!defined($condI) && !defined($condJ)) ||
				($attr[$attOrd{$condI}] && $attr[$attOrd{$condJ}])) {
				if ($l =~ /\+$/) {
					if ($needsReorder) {
						my @out;
						for (my $i = 0; $i < @features; $i++) {
							$out[$i] = $attr[$attOrd{$features[$i]}];
						}
						print POS join(',', @out) . ',+' . "\n";
					}
					else {
						print POS "$l\n";
					}
					$pos++;
				}
				else {
					if ($needsReorder) {
						my @out;
						for (my $i = 0; $i < @features; $i++) {
							$out[$i] = $attr[$attOrd{$features[$i]}];
						}
						print NEG join(',', @out) . ',-' . "\n";
					}
					else {
						print NEG "$l\n";
					}
					$neg++;
				}
			}
		}
		close (F);
		print "$file +: $pos, -:$neg\n";
	}
	close (POS);
	close (NEG);
}

# Selects the negative examples that have a distance less or equal than clustDist
sub ClusterData
{
	my ($model, $clustDist, $opt) = @_;
	my $datafile = $model->datasetFile;
	my $posFile = "$datafile.pos";
	
	&Attributes::Forget;
	
	open (F, "$datafile.negtots") || die "ooh";
	open (N, ">$datafile.neg") || die "ooh";
	open (Z, ">$datafile.neg0") || die "ooh";
	my $cont = 0;
	while (my $l = <F>) {
		chomp($l);
		my $dd;
		my @atributs = split(/\,/, $l);
		my $result = pop(@atributs);
		$dd = Attributes::MinDistance(\@atributs, $clustDist, $posFile);
		
		if ($dd == 0) {
			print Z "$l\n";
		}
		elsif ($dd <= $clustDist)  {
			print N "$l\n";
		}
		$cont++;
# 		if ($cont % 1000 == 0) {
# 			print "linies: $cont\n";
# 		}
	}
	close (N);
	close (Z);
	close (F);
	
# 	print "Treient contradictoris...\n";
# 	&Attributes::TreuContradictoris($outfile);
	
	# Afegim els negatius que queden
	system("cp $posFile $datafile");
	system("cat $datafile.neg >> $datafile");
	system("cat $datafile.neg0 >> $datafile");
	
	&Attributes::Forget;
}

# Train a (sub)model
sub TrainModel
{
	my ($model, $trainFiles, $step, $opt) = @_;
	
	# Conditions for 9-fold splitting
	my @condI = ('I_TYPE_P', 'I_TYPE_P', 'I_TYPE_P',
					 'I_TYPE_E', 'I_TYPE_E', 'I_TYPE_E',
					 'I_TYPE_S', 'I_TYPE_S', 'I_TYPE_S');
	my @condJ = ('J_TYPE_P', 'J_TYPE_E', 'J_TYPE_S',
					 'J_TYPE_P', 'J_TYPE_E', 'J_TYPE_S',
					 'J_TYPE_P', 'J_TYPE_E', 'J_TYPE_S');
	
	my $prfTree;
	if (defined($step)) {
		$prfTree = $model->modelFile . '-pr' . $step;
	}
	else {
		$prfTree = $model->modelFile;
	}
	
	my $mask;
	my $dt = new DT($prfTree);
	
	$dt->mask($model->mask);
	
	my $fData = $dt->datasetFile;
	system("rm $fData*") if (-e $fData);
	
	if (defined($step)) {
		SplitData($dt, $trainFiles, $condI[$step], $condJ[$step]);
	}
	else {
		SplitData($dt, $trainFiles);
	}
	
	my $dist = defined($opt->{dist}) ? $opt->{dist} : 4;
	ClusterData($dt, $dist);
	
	$dt->GenDT($opt);
	
	# Once we have the DT, we call c4.5rules to get the constraints
	my @constrStep = $dt->c45rules($opt);
# 	my @constrStep = $dt->ReadDTC45Rules("$prfTree.rulesoutput");
	if (defined($step)) {
		foreach my $c (@constrStep) {
			$c->[0] .= ' && $a{' . $condI[$step] . '} && $a{' . $condJ[$step] . '}';
		}
	}
	
	$dt->constraints(\@constrStep);
}

sub JoinSubmodels
{
	my ($model, $opt) = @_;
	
	# Put together all the constraints
	my $ccfile = $model->modelFile . '.constraints';
	system("rm $ccfile") if (-e $ccfile);
	for (my $step = 0; $step < 9; $step++) {
		my $prfTree = $model->modelFile . '-pr' . $step . '.constraints';
		system("cat $prfTree >> $ccfile");
	}
}


# Train a model spliting the data into independent parts
sub TrainModelSplit
{
	my ($model, $trainFiles, $opt) = @_;
	
	# Learn rules by using 9 different models
	for (my $step = 0; $step < 9; $step++) {
		TrainModel($model, $trainFiles, $step, $opt);
	}
	
	JoinSubmodels($model, $opt);
}



sub ComputeAccuracy
{
	my ($model, $trainFiles, $opt) = @_;
	my @pos;
	my @neg;
	my ($pcor, $tedges) = (0, 0);
	my $nConstraints = @{ $model->constraints };
	my $saveFile = $model->modelFile . '.acc';
	
	my $ii = 0;
	foreach my $file (@$trainFiles) {
		my $doc = new Document($file);
		my $elements = $doc->GetMentions($model->trueMentions, 1);
		
		if ($model->verbosity) {
			$ii++;
			print "\rComputing constraints accuracy... ($ii / " . scalar(@$trainFiles) . ") $file     ";
		}
		
		my $fName = "$file.constraints";
		$fName .= '-' . $model->alias if (defined($model->alias));
		if (open (F, $fName)) {
			while (my $l = <F>) {
				chomp($l);
				my ($edge, $lc) = split(/:/, $l);
				my ($indi, $indj) = split(/\,/, $edge);
				my $corefer = $doc->Corefer($elements->[$indi], $elements->[$indj]);
				$tedges++;
				$pcor++ if ($corefer);
				foreach my $c (split(/\,/, $lc)) {
					$pos[$c]++ if ($corefer);
					$neg[$c]++ if (!$corefer);
				}
			}
			close (F);
		}
		else {
			print "Warning: can't open $fName\n";
		}
			
		# free memory
		$elements = undef;
		$doc->Clear;
	}

	print "\n" if ($model->verbosity);
	open (MLOG, ">$saveFile");
	print MLOG "$pcor,$tedges\n";
	for (my $r = 0; $r < $nConstraints; $r++) {
		$pos[$r] = 0 if (!defined($pos[$r]));
		$neg[$r] = 0 if (!defined($neg[$r]));
		print MLOG "$pos[$r],$neg[$r]\n";
	}
	close (MLOG);
	
	return ($pcor, $tedges, \@pos, \@neg);
}

sub LoadAccuracy
{
	my ($model, $opt) = @_;
	my @pos;
	my @neg;
	my ($pcor, $tedges) = (0, 0);
	my $saveFile = $model->modelFile . '.acc';
	open (MLOG, $saveFile) || die "Can not open $saveFile";
	my $l = <MLOG>;
	chomp ($l);
	($pcor, $tedges) = split(/,/, $l);
	my $r = 0;
	while (my $l = <MLOG>) {
		chomp ($l);
		($pos[$r], $neg[$r]) = split(/,/, $l);
		$r++;
	}
	close (MLOG);
	
	return ($pcor, $tedges, \@pos, \@neg);
}

sub WeightConstraints
{
	my ($model, $opt) = @_;
	my @pes;
	my $nConstraints = @{ $model->constraints };
	my ($pcor, $tedges, $pos, $neg) = LoadAccuracy($model, $opt);
	
	# Precisio
	my $minAplica = defined($opt->{minAplica}) ? $opt->{minAplica} : 5;
	my $balance = defined($opt->{balance}) ? $opt->{balance} : 0.5;
	
	for (my $r = 0; $r < $nConstraints; $r++) {
		$pes[$r] = 0;
		my $aplica = $pos->[$r] + $neg->[$r];
		next if ($aplica < $minAplica || $aplica == 0);
		
		$pes[$r] = $pos->[$r] / $aplica - $balance;
	}
	
	@pes = nearest(0.0001, @pes);
	$model->weights(\@pes);
}


# Training process
sub Train
{
	my ($modelAlias, $trainFiles, $opt) = @_;
	
	# Create a model
	my $step = defined($opt->{stem}) ? $opt->{stem} : 'default';
	my $modelFile = "models/$modelAlias"; # TODO:$PATH_MODELS . '/' . $modelAlias . '/' . $stem;
	my $model = new Model($modelFile);
	$model->alias($modelAlias);
	
	TrainModelSplit($model, $trainFiles, $opt);
	
	# Apply constraints to the training files
	foreach my $file (@$trainFiles) {
		Prepare($model, $file, $opt);
	}
	
	# Calculate weights for constraints given training results:
	
	# - Compute accuracy
	ComputeAccuracy($model, $trainFiles, $opt);

	# - default balance: 0.5
	WeightConstraints($model, $opt);
}


# sub ApplyConstraintsElem
# {
# 	my ($model, $data, $nmentions) = @_;
# 	my %apply;
# 	my $constraints = $model->constraints;
# 	my @features = $model->features;
# 	
# 	# Per a que s'executi mes rapid, faig nomes un "eval"
# 	my $code = '
# 	my $index = 0;
# 	for (my $i = 0; $i < $nmentions; $i++) {
# 		for (my $j = $i+1; $j < $nmentions; $j++) {
# 			my %a;
# 			for (my $k = 0; $k < @features; $k++) {
# 				if (defined($data->[$index][$k])) {
# 					$a{$features[$k]} = $data->[$index][$k];
# 				}
# 				else {
# 					die "i: $i, j: $j, k: $k Error !!";
# 				}
# 			}
# 			
# 			my $r;
# 			my $edge = "$i,$j";
# ';
# 
# 	for (my $k = 0; $k < scalar(@$constraints); $k++) {
# 		my ($cond, $val) = @{$constraints->[$k]};
# 		$code .= $cond . ";\n";
# 		$code .= 'if ($r) {
# 	push(@{$apply{$edge}}, ' . $k . ');
# }
# ';
# 	}
# 	$code .= '$index++;
# 		}}';
# 	eval $code;
# 	
# 	return \%apply;
# }
# 
# # Apply the constraints to a file (using cache .data)
# sub Prepare
# {
# 	my ($model, $fitxer) = @_;
# 	
# 	open (N, "$fitxer.nm") || die "No trobo $fitxer.nm";
# 	my $nmentions = <N>;
# 	chomp($nmentions);
# 	close (N);
# 	print "$nmentions mentions\n" if ($model->verbosity);
# 	
# 	my $data = LoadData($model, $fitxer);
# 	my $apply = ApplyConstraintsElem($model, $data, $nmentions);
# 	
# 	my $outfile = "$fitxer.constraints";
# 	if (defined( $model->alias )) {
# 		$outfile .= '-' . $model->alias;
# 	}
# 	
# 	open (G, ">$outfile") || die "Can not open $outfile";
# 	foreach my $edge (keys(%{$apply})) {
# 		print G "$edge:" . join(',', @{$apply->{$edge}}) . "\n";
# 	}
# 	close (G);
# 	$data = undef;
# 	$apply = undef;
# }


sub ReduceCostCode
{
	my ($A, $B, $DATA) = @_;
	
	my $reduceCost;
	
	if ($A eq 'ini') {
		$reduceCost = '
		my $sk = 0;
		my @skip = ();
';
	}
	elsif ($A eq 'ini_lvl') {
		$reduceCost = '
		my @lev' . $B . ' = ();
';
	}
	else {
		$reduceCost = '
		# Reduce iterations by precalculating distance features
		my %b;
		for (my $kk = 0; $kk < @pairFeatures; $kk++) {
			$b{$pairFeatures[$kk]} = ' . $DATA . '[$attOrdPair{$pairFeatures[$kk]}];
		}
		
		for (my $cc = 0; $cc < scalar(@$constraints); $cc++) {
			if (!$skip[$cc]) {
				my ($cond, $val) = @{$constraints->[$cc]};
				while ($cond =~ s/\{DIST_SEN_(L?\d)_(\d)(\d)\}\s*(\&\&|\=\=\s*1|$)/ /) {
					my ($dist, $ni, $nj) = ($1, $2, $3);
					if ($ni == ' .  $A . ' && $nj == ' . $B . ') {
					if ( ($dist eq \'0\' && $b{DIST_SEN_0} == 0) ||
						  ($dist eq \'1\' && $b{DIST_SEN_0} == 0 && $b{DIST_SEN_1} == 0) ||
						  ($dist eq \'L3\' && $b{DIST_SEN_L3} == 0) ) {
						$skip[$cc] = 1;
						$sk++;
						push(@lev' . $B . ', $cc);
						last;
					}
					}
				}
			}
		}
		if ($sk >= scalar(@$constraints)) {
			foreach my $cc (@lev' . $B . ') {
				$skip[$cc] = 0;
				$sk--;
			}
			last;
		}
		$b = ();
';
	}
	return $reduceCost;
# 	return '';
}


# Apply the constraints to a file (using cache .data)
sub Prepare
{
	my ($model, $file, $Nmax) = @_;
	my @features = $model->features;
	my @pairFeatures;
	if (defined($Nmax) && $Nmax > 2) {
		@features = $model->groupFeatures($Nmax);
		@pairFeatures = $model->features;
	}
	
	open (N, "$file.nm") || die "Error opening $file.nm";
	my $nmentions = <N>;
	chomp($nmentions);
	close (N);
	
	my $outfile = "$file.constraints";
	$outfile .= '-' . $model->alias if (defined( $model->alias ));
	open (G, ">$outfile") || die "Can not open $outfile";

	my %attOrd;
	my $needsReorder = CheckFeatures($model, $file, \%attOrd, $Nmax);
	my %attOrdPair;
	CheckFeatures($model, $file, \%attOrdPair);
	
	my $fileHand;
	open ($fileHand, "$file.data") || die "oops $file.data: $!";
	my $attrline = <$fileHand>; # first line
	my $offset = tell($fileHand);
	chomp($attrline);
	my $nfeat = split(/\,/, $attrline);
	
	# read another line in order to know if last value is the class
	my $incClass = 0;
	my $line = <$fileHand>; # first line
	chomp($line);
	if ($line =~ /\,[\+\-]$/) {
		$incClass = 1;
	}
	
	my $code;
	if (!defined($Nmax) || $Nmax == 2) {
		$code = '
my $index = 0;
for (my $i = 0; $i < $nmentions; $i++) {
	for (my $j = $i+1; $j < $nmentions; $j++) {
		my $edge = "$i,$j";
		my %a;
		my @data = GetPairVector($fileHand, $i, $j, $nmentions, $nfeat, $offset, $incClass);
';
	}
	elsif ($Nmax == 3) {
		$code = '
		my $index = 0;
		for (my $i = 0; $i < $nmentions; $i++) {
			for (my $j = $i+1; $j < $nmentions; $j++) {
				my @data_ij = GetPairVector($fileHand, $i, $j, $nmentions, $nfeat, $offset, $incClass);
				' . ReduceCostCode('ini') . '
				' . ReduceCostCode('ini_lvl', 1) . '
				' . ReduceCostCode('0', '1', '$data_ij') . '
				
				for (my $k = $j+1; $k < $nmentions; $k++) {
					my @data_ik = GetPairVector($fileHand, $i, $k, $nmentions, $nfeat, $offset, $incClass);
					my @data_jk = GetPairVector($fileHand, $j, $k, $nmentions, $nfeat, $offset, $incClass);
					' . ReduceCostCode('ini_lvl', 2) . '
					' . ReduceCostCode('0', '2', '$data_ik') . '
					' . ReduceCostCode('1', '2', '$data_jk') . '
					
					my %a;
					my @data = (@data_ij, @data_ik, @data_jk);
					my $edge = "$i,$j,$k";
';
	}
	elsif ($Nmax == 4) {
		
		

		$code = '
		my $index = 0;
		for (my $i = 0; $i < $nmentions; $i++) {
			
			for (my $j = $i+1; $j < $nmentions; $j++) {
				my @data_ij = GetPairVector($fileHand, $i, $j, $nmentions, $nfeat, $offset, $incClass);
				' . ReduceCostCode('ini') . '
				' . ReduceCostCode('ini_lvl', 1) . '
				' . ReduceCostCode('0', '1', '$data_ij') . '
				
				for (my $k = $j+1; $k < $nmentions; $k++) {
					my @data_ik = GetPairVector($fileHand, $i, $k, $nmentions, $nfeat, $offset, $incClass);
					my @data_jk = GetPairVector($fileHand, $j, $k, $nmentions, $nfeat, $offset, $incClass);
					' . ReduceCostCode('ini_lvl', 2) . '
					' . ReduceCostCode('0', '2', '$data_ik') . '
					' . ReduceCostCode('1', '2', '$data_jk') . '
					
					for (my $l = $k+1; $l < $nmentions; $l++) {
						my @data_il = GetPairVector($fileHand, $i, $l, $nmentions, $nfeat, $offset, $incClass);
						my @data_jl = GetPairVector($fileHand, $j, $l, $nmentions, $nfeat, $offset, $incClass);
						my @data_kl = GetPairVector($fileHand, $k, $l, $nmentions, $nfeat, $offset, $incClass);
						' . ReduceCostCode('ini_lvl', 3) . '
						' . ReduceCostCode('0', '3', '$data_il') . '
						' . ReduceCostCode('1', '3', '$data_jl') . '
						' . ReduceCostCode('2', '3', '$data_kl') . '
						
						my %a;
						my @data = (@data_ij, @data_ik, @data_il, @data_jk, @data_jl, @data_kl);
						my $edge = "$i,$j,$k,$l";
';
	}
	
	$code .= '
		if ($needsReorder) {
			for (my $kk = 0; $kk < @features; $kk++) {
				$a{$features[$kk]} = $data[$attOrd{$features[$kk]}];
				#if ($i == 476 && $j == 478 && $k == 481 && $l == 483) {
				#	print "$features[$kk] ($attOrd{$features[$kk]}) = $a{$features[$kk]}\n";
				#}
			}
		}
		else {
			for (my $kk = 0; $kk < @features; $kk++) {
				$a{$features[$kk]} = $data[$kk];
			}
		}
		
		my $r;
		my @apply;
';

	my $constraints = $model->constraints;
	for (my $k = 0; $k < scalar(@$constraints); $k++) {
		my ($cond, $val) = @{$constraints->[$k]};
		$code .= $cond . ";\n";
		$code .= 'if ($r) {
	push(@apply, ' . $k . ');
}
';
	}
	if (!defined($Nmax) || $Nmax == 2) {
		$code .= 'print G "$edge:" . join(",", @apply) . "\n";
	$index++;
}}';
	}
	else {
		$code .= 'print G "$edge:" . join(",", @apply) . "\n" if (@apply);
		print "$edge:" . join(",", @apply) . "\n" if (@apply);
		$index++;
		';
		if ($Nmax == 3) {
			$code .= '}}}';
		}
		elsif ($Nmax == 4) {
			$code .= '}}}}';
		}
		
	}
	
	
	eval $code;
	if ($@) {
		print "CODE>>\n$code\n<<CODE\n";
# 		print $@;
		die;
	}
	close (G);
	close (F);
}

# gets the vector of features of a pair of mentions from the .data file
sub GetPairVector
{
	my ($fileHand, $mi, $mj, $nmentions, $nfeat, $offset, $teClass) = @_;
	
	# Calculate the line of the text file that we need
	my $line = $mi * $nmentions - ($mi + 1) * $mi / 2 + ($mj - $mi) - 1; # first line = 0
	
	my $nbytesline = $nfeat * 2; #$nfeat values + $nfeat-1 commas + 1 \n
	if ($teClass) {
		$nbytesline += 2;
	}
	
	seek($fileHand, $offset + ($line * $nbytesline), 0);
	my $l = <$fileHand>;
	chomp($l);
	my @atr = split(/\,/, $l);
	if ($teClass) {
		pop(@atr);
	}
	return @atr;
}


# Apply group constraint
# Nmax = maximum number of mentions involved
# sub PrepareN
# {
# 	my ($model, $file, $Nmax) = @_;
# 	my @features = $model->groupFeatures($Nmax);
# 	
# 	open (N, "$file.nm") || die "Error opening $file.nm";
# 	my $nmentions = <N>;
# 	chomp($nmentions);
# 	close (N);
# 	
# 	my $outfile = "$file.constraints";
# 	$outfile .= '-' . $model->alias if (defined( $model->alias ));
# 	open (G, ">$outfile") || die "Can not open $outfile";
# 
# 	my %attOrd;
# 	my $needsReorder = CheckFeatures($model, $file, \%attOrd, $Nmax);
# 	my $fileHand;
# 	open ($fileHand, "$file.data") || die "oops $file.data: $!";
# 	my $attrline = <$fileHand>; # first line
# 	my $offset = tell($fileHand);
# 	my $nfeat = split(/\,/, $attrline);
# 	
# 	if ($Nmax == 3) {
# 		my $code = '
# 		my $index = 0;
# 		for (my $i = 0; $i < $nmentions; $i++) {
# 			for (my $j = $i+1; $j < $nmentions; $j++) {
# 				my @data_ij = GetPairVector($fileHand, $i, $j, $nmentions, $nfeat);
# 				for (my $k = $j+1; $k < $nmentions; $k++) {
# 					my @data_ik = GetPairVector($fileHand, $i, $k, $nmentions, $nfeat);
# 					my @data_jk = GetPairVector($fileHand, $j, $k, $nmentions, $nfeat);
# 					my %a;
# 					my @data = (@data_ij, @data_ik, @data_jk);
# 					for (my $k = 0; $k < @features; $k++) {
# 						$a{$features[$k]} = $data[$attOrd{$features[$k]}];
# 					}
# 				
# 					my $r;
# 					my $edge = "$i,$j,$k";
# 					my @apply;
# ';
# 	}
# 	elsif ($Nmax == 4) {
# 		my $code = '
# 		my $index = 0;
# 		for (my $i = 0; $i < $nmentions; $i++) {
# 			for (my $j = $i+1; $j < $nmentions; $j++) {
# 				my @data_ij = GetPairVector($fileHand, $i, $j, $nmentions, $nfeat);
# 				for (my $k = $j+1; $k < $nmentions; $k++) {
# 					my @data_ik = GetPairVector($fileHand, $i, $k, $nmentions, $nfeat);
# 					my @data_jk = GetPairVector($fileHand, $j, $k, $nmentions, $nfeat);
# 					for (my $l = $k+1; $l < $nmentions; $l++) {
# 						my @data_il = GetPairVector($fileHand, $i, $l, $nmentions, $nfeat);
# 						my @data_jl = GetPairVector($fileHand, $j, $l, $nmentions, $nfeat);
# 						my @data_kl = GetPairVector($fileHand, $k, $l, $nmentions, $nfeat);
# 						my %a;
# 						my @data = (@data_ij, @data_ik, @data_jk);
# 						for (my $k = 0; $k < @features; $k++) {
# 							$a{$features[$k]} = $data[$attOrd{$features[$k]}];
# 						}
# 				
# 						my $r;
# 						my $edge = "$i,$j,$k,$l";
# 						my @apply;
# ';
# 	}
# 	my $constraints = $model->constraints;
# 	for (my $k = 0; $k < scalar(@$constraints); $k++) {
# 		my ($cond, $val) = @{$constraints->[$k]};
# 		$code .= $cond . ";\n";
# 		$code .= 'if ($r) {
# 	push(@apply, ' . $k . ');
# }
# ';
# 	}
# 	$code .= 'print G "$edge:" . join(",", @apply) . "\n" if (@apply);
# 	$index++;
# 		}}';
# 	if ($Nmax == 3) {
# 		$code .= '}';
# 	}
# 	elsif ($Nmax == 4) {
# 		$code .= '}}';
# 	}
# 	
# 	eval $code;
# 	if ($@) {
# 		print "CODE>>\n$code\n<<CODE\n";
# # 		print $@;
# 		die;
# 	}
# 	close (G);
# 	close (F);
# }

sub Solve
{
	my ($model, $doc, $opt) = @_;
	
	my $solver = new SolverRelax($model);
	$solver->verbosity($model->verbosity);
	$solver->Solve($doc, $opt);
}

1;
