#!/usr/bin/perl

use FindBin qw( $RealBin );
use lib "$RealBin/lib";

use strict;
use Getopt::Long;
use RelaxCor;
use MentionDetection;
use Document;
use CorScorer;
use POSIX ":sys_wait_h"; # fork

###############################################################################
# globals

use Globals qw ($PATH_HOME $PATH_MODELS $STEM);

my $VERBOSE = 0;
my $trueMentions = 0;

my $DEFAULT_SECTION = 'sec-default';
my $DEFAULT_MODEL = 'default';

###############################################################################
# functions




###############################################################################
# main

my $result;
my %opt;
if (@ARGV) {
	$result = GetOptions (\%opt, 'model=s', 'section=s', 'verbose', 'N=i',
								 'parallel=i', 'balance=s', 'measure=s', 'emodel=s@', 'reduceMentions');
}

if (!$result) {
	print '
us: run-coref.pl [options] <actions> <file> [more files..]

Options:
	--model=<name>
		Name of the model

	--section=<name>
		Name of the section or corpora (ontonotes, ace, ...)

	--N=<#>
		Maximum number of edges per vertex (pruning) (2, 4, 6..)

	--balance=<#>
		Balance value, between 0 and 1

	--parallel=<#>
		Number of parallel executions

	--measure=<name>
		If action \'e\' is specified, the corpora is evaluated using this measure.
		Choices: muc, ceafe, ceafm, bcub, blanc

	--verbose
		More information in the output.
	

Actions:

	d	Detect mentions
	g	Generate features far all mention pairs
	a	Apply constraints of the model
	s	Solve coreferences
	e	Evaluate with a measure
';
	exit;
}

my $accions = shift(@ARGV);

if ($accions =~ /[^dgase]/) {
	print "Error: unknown actions ($accions)\n";
	exit;
}

my $modelAlias = defined($opt{model}) ? $opt{model} : $DEFAULT_MODEL;
my $section = defined($opt{section}) ? $opt{section} : $DEFAULT_SECTION;
my $modelFile = "$PATH_HOME/models/$modelAlias/$STEM-$section";

my $model = new Model($modelFile);
$model->alias($modelAlias);
$model->verbosity($VERBOSE);
$model->trueMentions($trueMentions);

# extra models?
if (defined($opt{emodel}) && @{$opt{emodel}}) {
	@{$opt{extraModels}} = ();
	foreach my $extraModelName (@{$opt{emodel}}) {
		my $eModelFile = "$PATH_HOME/models/$extraModelName/$STEM-$section";
		my $eModel = new Graph($eModelFile);
		$eModel->alias($extraModelName);
		$eModel->verbosity($VERBOSE);
		$eModel->trueMentions($trueMentions);
		push(@{$opt{extraModels}}, $eModel);
		print "Entity-mention model: $extraModelName\n";
	}
}


if (defined($opt{balance})) {
	&RelaxCor::WeightConstraints($model, \%opt);
}

if (defined($opt{N})) {
	$opt{Nprune} = $opt{N};
}

my @files = @ARGV;
if ($accions =~ /[dgas]/) {
	my $ii = 0;
	my $kids = 0;
	foreach my $file (@files) {
		$ii++;
		print "($ii\/" . scalar(@files) . ") $file \n";
		
		my $pid = -1;
		if (defined($opt{parallel})) {
			$pid = fork;
			if (!defined($pid)) {
				die "Error forking $!";
			}
		}
		
		if ($pid <= 0) { # child or unique process
			my $doc = new Document($file);
			
			# Detectar mencions
			if ($accions =~ /d/) {
				&MentionDetection::Detect( $doc, \%opt);
			}
			
			# Generar atributs per totes les parelles de mencions
			if ($accions =~ /g/) {
				&RelaxCor::GenData($model, $doc);
			}
			
			# Aplicar les restriccions del model
			if ($accions =~ /a/) {
				&RelaxCor::Prepare($model, $file);
			}
			
			# Calcular coreferencies. Relax
			if ($accions =~ /s/) {
				&RelaxCor::Solve($model, $doc, \%opt);
			}
			
			$doc->Clear;
			
			if ($pid == 0) { # child
				exit(0);
			}
		}
		
		
		if ($pid > 0) { # main process with kids
			$kids++;
			
			# don't forget zombies
			my $child;
			do {
				$child = waitpid(-1, WNOHANG);
				$kids-- if ($child > 0);
			} while ($child > 0);
			
			if ($kids >= $opt{parallel}) {
				wait; # waits 1 process
				$kids--;
			}
		}
	}

	if (defined($opt{parallel})) {
		while ($kids > 0) {
			wait;
			$kids--;
		}
	}
}

if ($accions =~ /e/) {
	my $metric;
	if (defined($opt{measure})) {
		$metric = $opt{measure};
	}
	else {
		print "Evaluating with default measure: muc\n";
		$metric = 'muc';
	}
	
	my ($acumNR, $acumDR, $acumNP, $acumDP) = (0, 0, 0, 0);

	# We assume that "$file.gold" exists for each $file
	foreach my $file (@files) {
		my $fileOut = ($metric eq 'det') ? $file : "$file.out";
		my ($dNR, $dDR, $dNP, $dDP) = &CorScorer::Score( $metric, "$file.gold", $fileOut, '__single');
		$acumNR += $dNR;
		$acumDR += $dDR;
		$acumNP += $dNP;
		$acumDP += $dDP;
	}
	
	my $p = ($acumDP) ? $acumNP / $acumDP : 0;
	my $r = ($acumDR) ? $acumNR / $acumDR : 0;
	my $f1 = ($p + $r) ? 2 * $p * $r / ($p + $r) : 0;
	print "\n\n$metric ==>\tPRE: $p, REC: $r, F1: $f1\n";
}
