#!/usr/bin/perl

use FindBin qw( $RealBin );
use lib "$RealBin/lib";

use strict;
use Getopt::Long;
use RelaxCor;
use MentionDetection;
use Document;

###############################################################################
# globals

use Globals qw ($PATH_HOME $PATH_MODELS $STEM);

my $VERBOSE = 0;

my $DEFAULT_SECTION = 'sec-default';
my $DEFAULT_MODEL = 'default';


###############################################################################
# functions




###############################################################################
# main

my $result;
my %opt;
if (@ARGV) {
	$result = GetOptions (\%opt, 'model=s', 'verbose', 'step=i', 'section=s',
								 'trueMentions', 'dist=i', 'Ngroup=i', 'reduceMentions');
}

if (!$result) {
	print '
us: ./run-train.pl [options] <action(s)> <file> [more files..]

Options:
	--model=<NAME>
		Name of the model

	--section=<NAME>
		Name of the section or corpora (ontonotes, ace, ...)

	--step=<#>
		The training process can be divided into 9+1 independent steps. This option
		executes the indicated step. (step 9 requires the previous execution of steps 0 to 8)

	--verbose
		More information in the output.
	

Actions:

	d	Detect mentions
	g	Generate features far all mention pairs
	t	Train a model
	a	Apply constraints of the model
';
	exit;
}

my $accions = shift(@ARGV);

if ($accions =~ /[^dgat]/) {
	print "Error: unknown actions ($accions)\n";
	exit;
}

my $modelAlias = defined($opt{model}) ? $opt{model} : $DEFAULT_MODEL;
my $section = defined($opt{section}) ? $opt{section} : $DEFAULT_SECTION;
my $modelFile = "$PATH_HOME/models/$modelAlias/$STEM-$section";
my $trueMentions = defined($opt{trueMentions}) ? 1 : 0;


my $model = new Model($modelFile);
$model->alias($modelAlias);
$model->verbosity($VERBOSE);
$model->trueMentions($trueMentions);

my ($detok, $detpre, $detrec) = (0, 0, 0);
my @files = @ARGV;
my $ii = 0;
if ($accions =~ /[dga]/) {
	foreach my $file (@files) {
		$ii++;
		print "($ii\/" . scalar(@files) . ") $file \n";
		my $doc = new Document($file);
		
		# Detectar mencions
		if ($accions =~ /d/) {
			
			# We assume that current file has gold mention annotations
			if (!-e "$file.gold") {
				system("cp $file $file.gold");
			}
			&MentionDetection::Detect( $doc, \%opt);
			
			my $docGold = new Document("$file.gold");
			my ($a, $b, $c) = &MentionDetection::AnnotateGold( $doc, $docGold);
			$detok += $a;
			$detpre += $b;
			$detrec += $c;
		}
		elsif ($trueMentions && !-e "$file.gold") {
			system("cp $file $file.gold");
		}
		
		# Generar atributs per totes les parelles de mencions
		if ($accions =~ /g/) {
			
			&RelaxCor::GenData($model, $doc, 1);
		}
		
		# Aplicar les restriccions del model
		if ($accions =~ /a/) {
			&RelaxCor::Prepare($model, $file, $opt{Ngroup});
		}
		
		$doc->Clear;
	}
}

if ($accions =~ /d/) {
	my $pre = $detpre ? $detok / $detpre : 0;
	my $rec = $detrec ? $detok / $detrec : 0;
	my $f1 = ($rec + $pre) ? 2*$rec*$pre/($rec+$pre) : 0;
	print "Mention detection: P: $pre, R: $rec, F1: $f1\n";
}


if ($accions =~ /t/) {
	
	# Use a half of the files to generate the constraints
	my @train;
	for (my $i = 0; $i < @files; $i += 1) {
		push(@train, $files[$i]) if ($i % 4 == 0); #25%
	}
	
	if (!defined($opt{step})) {
		# Train en 1 execucio
		&RelaxCor::TrainModelSplit($model, \@train, \%opt);
	}
	elsif ($opt{step} < 9) {
		&RelaxCor::TrainModel($model, \@train, $opt{step}, \%opt);
	}
	
	if (!defined($opt{step}) || $opt{step} == 9) {
		
		if ($opt{step} == 9) {
			&RelaxCor::JoinSubmodels($model, \%opt);
		}
		
		# Apply constraints to the training files
		for (my $i = 0; $i < @files; $i++) {
			print "($i\/" . scalar(@files) . ") Applying constraints to $files[$i] \n";
			&RelaxCor::Prepare($model, $files[$i], \%opt);
		}
		
		# - Compute accuracy
		print "Computing accuracy...\n";
		&RelaxCor::ComputeAccuracy($model, \@files, \%opt);

		# - default balance: 0.5
		&RelaxCor::WeightConstraints($model, \%opt);
	}
}

